# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import json
import yaml
import tempfile
import jsonpickle
import datetime
from dataclasses import dataclass, field

from .models import (
    Load,
    LoadType,
    ObjectList,
    TaskCall,
    TaskCallsInTree,
    AnsibleRunContext,
    ARIResult,
)
from .loader import (
    get_loader_version,
)
from .parser import Parser
from .model_loader import load_object
from .tree import TreeLoader
from .annotators.variable_resolver import resolve_variables
from .analyzer import analyze
from .risk_detector import detect
from .dependency_dir_preparator import (
    DependencyDirPreparator,
)
from .findings import Findings
from .risk_assessment_model import RAMClient
import ansible_risk_insight.logger as logger
from .utils import (
    is_url,
    is_local_path,
    escape_url,
    escape_local_path,
    summarize_findings,
    summarize_findings_data,
    split_target_playbook_fullpath,
    split_target_taskfile_fullpath,
    equal,
)


default_config_path = os.path.expanduser("~/.ari/config")
default_data_dir = os.path.join("/tmp", "ari-data")
default_rules_dir = os.path.join(os.path.dirname(__file__), "rules")
default_log_level = "info"
default_rules = []
default_disable_default_rules = False
default_logger_key = "ari"


@dataclass
class Config:
    path: str = ""

    data_dir: str = ""
    rules_dir: str = ""
    logger_key: str = ""
    log_level: str = ""
    rules: list = field(default_factory=list)
    disable_default_rules: bool = False

    _data: dict = field(default_factory=dict)

    def __post_init__(self):
        if not self.path:
            self.path = default_config_path
        config_data = {}
        if os.path.exists(self.path):
            with open(self.path, "r") as file:
                try:
                    config_data = yaml.safe_load(file)
                except Exception as e:
                    raise ValueError(f"failed to load the config file: {e}")
        self._data = config_data

        if not self.data_dir:
            self.data_dir = self._get_single_config("ARI_DATA_DIR", "data_dir", default_data_dir)
        if not self.disable_default_rules:
            self.disable_default_rules = self._get_single_config("ARI_DISABLE_DEFAULT_RULES", "disable_default_rules", default_disable_default_rules)
        if not self.rules_dir:
            self.rules_dir = self._get_single_config("ARI_RULES_DIR", "rules_dir", default_rules_dir)
        # automatically add the default rules dir unless it is disabled
        if not self.rules_dir.endswith(default_rules_dir) and not self.disable_default_rules:
            self.rules_dir += ":" + default_rules_dir
        if not self.logger_key:
            self.logger_key = self._get_single_config("ARI_LOGGER_KEY", "logger_key", default_logger_key)
        if not self.log_level:
            self.log_level = self._get_single_config("ARI_LOG_LEVEL", "log_level", default_log_level)
        if not self.rules:
            self.rules = self._get_single_config("ARI_RULES", "rules", default_rules, "list", ",")

    def _get_single_config(self, env_key: str = "", yaml_key: str = "", __default: any = None, __type=None, separator=""):
        _from_env = os.environ.get(env_key, None)
        if _from_env and __type:
            if __type == "list":
                _from_env = _from_env.split(separator)
        _from_file = self._data.get(yaml_key, None)
        value = _from_env or _from_file or __default
        return value


collection_manifest_json = "MANIFEST.json"
role_meta_main_yml = "meta/main.yml"
role_meta_main_yaml = "meta/main.yaml"


supported_target_types = [
    LoadType.PROJECT,
    LoadType.COLLECTION,
    LoadType.ROLE,
    LoadType.PLAYBOOK,
]

config = Config()

logger.set_logger_channel(config.logger_key)
logger.set_log_level(config.log_level)


@dataclass
class SingleScan(object):
    type: str = ""
    name: str = ""
    collection_name: str = ""
    role_name: str = ""
    target_playbook_name: str = None
    playbook_yaml: str = ""
    playbook_only: bool = False
    target_taskfile_name: str = None
    taskfile_yaml: str = ""
    taskfile_only: bool = False

    skip_playbook_format_error: bool = (True,)
    skip_task_format_error: bool = (True,)

    install_log: str = ""
    tmp_install_dir: tempfile.TemporaryDirectory = None

    index: dict = field(default_factory=dict)

    root_definitions: dict = field(default_factory=dict)
    ext_definitions: dict = field(default_factory=dict)

    trees: list = field(default_factory=list)
    # for inventory object
    additional: ObjectList = field(default_factory=ObjectList)

    taskcalls_in_trees: list = field(default_factory=list)
    contexts: list = field(default_factory=list)

    data_report: dict = field(default_factory=dict)

    __path_mappings: dict = field(default_factory=dict)

    install_dependencies: bool = False

    dependency_dir: str = ""
    target_path: str = ""
    loaded_dependency_dirs: list = field(default_factory=list)
    use_src_cache: bool = True

    prm: dict = field(default_factory=dict)

    download_url: str = ""
    version: str = ""
    hash: str = ""

    source_repository: str = ""
    out_dir: str = ""

    extra_requirements: list = field(default_factory=list)
    resolve_failures: dict = field(default_factory=dict)

    findings: Findings = None
    result: ARIResult = None

    # the following are set by ARIScanner
    root_dir: str = ""
    rules_dir: str = ""
    rules: list = field(default_factory=list)
    spec_mutations_from_previous_scan: dict = field(default_factory=dict)
    spec_mutations: dict = field(default_factory=dict)
    use_ansible_doc: bool = True
    do_save: bool = False
    silent: bool = False
    _parser: Parser = None

    def __post_init__(self):
        if self.type == LoadType.COLLECTION or self.type == LoadType.ROLE:
            type_root = self.type + "s"
            target_name = self.name
            if is_local_path(target_name):
                target_name = escape_local_path(target_name)
            self.__path_mappings = {
                "src": os.path.join(self.root_dir, type_root, "src"),
                "root_definitions": os.path.join(
                    self.root_dir,
                    type_root,
                    "root",
                    "definitions",
                    type_root,
                    target_name,
                ),
                "ext_definitions": {
                    LoadType.ROLE: os.path.join(self.root_dir, "roles", "definitions"),
                    LoadType.COLLECTION: os.path.join(self.root_dir, "collections", "definitions"),
                },
                "index": os.path.join(
                    self.root_dir,
                    type_root,
                    "{}-{}-index-ext.json".format(self.type, target_name),
                ),
                "install_log": os.path.join(
                    self.root_dir,
                    type_root,
                    "{}-{}-install.log".format(self.type, target_name),
                ),
            }

        elif self.type == LoadType.PROJECT or self.type == LoadType.PLAYBOOK or self.type == LoadType.TASKFILE:
            type_root = self.type + "s"
            proj_name = escape_url(self.name)
            if self.type == LoadType.PLAYBOOK:
                if self.playbook_yaml:
                    self.target_playbook_name = self.name
                else:
                    _, self.target_playbook_name = split_target_playbook_fullpath(self.name)
            elif self.type == LoadType.TASKFILE:
                if self.taskfile_yaml:
                    self.target_taskfile_name = self.name
                else:
                    _, self.target_taskfile_name = split_target_taskfile_fullpath(self.name)
            self.__path_mappings = {
                "src": os.path.join(self.root_dir, type_root, proj_name, "src"),
                "root_definitions": os.path.join(
                    self.root_dir,
                    type_root,
                    proj_name,
                    "definitions",
                ),
                "ext_definitions": {
                    LoadType.ROLE: os.path.join(self.root_dir, "roles", "definitions"),
                    LoadType.COLLECTION: os.path.join(self.root_dir, "collections", "definitions"),
                },
                "index": os.path.join(
                    self.root_dir,
                    type_root,
                    proj_name,
                    "index-ext.json",
                ),
                "install_log": os.path.join(
                    self.root_dir,
                    type_root,
                    proj_name,
                    "{}-{}-install.log".format(self.type, proj_name),
                ),
                "dependencies": os.path.join(self.root_dir, type_root, proj_name, "dependencies"),
            }

        else:
            raise ValueError("Unsupported type: {}".format(self.type))

        if self.playbook_yaml:
            self.playbook_only = True
            if not self.name:
                self.name = "__in_memory__"
                self.target_playbook_name = self.name

        if self.taskfile_yaml:
            self.taskfile_only = True
            if not self.name:
                self.name = "__in_memory__"
                self.target_taskfile_name = self.name

    def make_target_path(self, typ, target_name, dep_dir=""):
        target_path = ""

        if dep_dir:
            parts = target_name.split(".")
            if len(parts) == 1:
                parts.append("")
            dep_dir_target_path_candidates = [
                os.path.join(dep_dir, target_name),
                os.path.join(dep_dir, parts[0], parts[1]),
                os.path.join(dep_dir, "ansible_collections", parts[0], parts[1]),
            ]
            for cand_path in dep_dir_target_path_candidates:
                if os.path.exists(cand_path):
                    target_path = cand_path
                    break
        if target_path != "":
            return target_path

        if typ == LoadType.COLLECTION:
            parts = target_name.split(".")
            if is_local_path(target_name):
                target_path = target_name
            else:
                target_path = os.path.join(self.root_dir, typ + "s", "src", "ansible_collections", parts[0], parts[1])
        elif typ == LoadType.ROLE:
            if is_local_path(target_name):
                target_path = target_name
            else:
                target_path = os.path.join(self.root_dir, typ + "s", "src", target_name)
        elif typ == LoadType.PROJECT:
            if is_url(target_name):
                target_path = os.path.join(self.get_src_root(), escape_url(target_name))
            else:
                target_path = target_name
        elif typ == LoadType.PLAYBOOK:
            if is_url(target_name):
                target_path = os.path.join(self.get_src_root(), escape_url(target_name))
            else:
                target_path = target_name
        elif typ == LoadType.TASKFILE:
            if is_url(target_name):
                target_path = os.path.join(self.get_src_root(), escape_url(target_name))
            else:
                target_path = target_name
        return target_path

    def get_src_root(self):
        return self.__path_mappings["src"]

    def is_src_installed(self):
        index_location = self.__path_mappings["index"]
        return os.path.exists(index_location)

    def _prepare_dependencies(self, root_install=True):
        # Install the target if needed
        target_path = self.make_target_path(self.type, self.name)

        # Dependency Dir Preparator
        ddp = DependencyDirPreparator(
            root_dir=self.root_dir,
            source_repository=self.source_repository,
            target_type=self.type,
            target_name=self.name,
            target_version=self.version,
            target_path=target_path,
            target_dependency_dir=self.dependency_dir,
            target_path_mappings=self.__path_mappings,
            do_save=self.do_save,
            silent=self.silent,
            tmp_install_dir=self.tmp_install_dir,
        )
        dep_dirs = ddp.prepare_dir(
            root_install=root_install,
            is_src_installed=self.is_src_installed(),
            cache_enabled=self.use_src_cache,
            cache_dir=os.path.join(self.root_dir, "archives"),
        )

        self.target_path = target_path
        self.version = ddp.metadata.version
        self.hash = ddp.metadata.hash
        self.download_url = ddp.metadata.download_url
        self.loaded_dependency_dirs = dep_dirs

        return target_path, dep_dirs

    def create_load_file(self, target_type, target_name, target_path):

        loader_version = get_loader_version()

        if not os.path.exists(target_path) and not self.playbook_yaml and not self.taskfile_yaml:
            raise ValueError("No such file or directory: {}".format(target_path))
        if not self.silent:
            logger.debug(f"target_name: {target_name}")
            logger.debug(f"target_type: {target_type}")
            logger.debug(f"path: {target_path}")
            logger.debug(f"loader_version: {loader_version}")
        ld = Load(
            target_name=target_name,
            target_type=target_type,
            path=target_path,
            loader_version=loader_version,
            playbook_yaml=self.playbook_yaml,
            playbook_only=self.playbook_only,
            taskfile_yaml=self.taskfile_yaml,
            taskfile_only=self.taskfile_only,
        )
        load_object(ld)
        return ld

    def get_definition_path(self, ext_type, ext_name):
        target_path = ""
        if ext_type == LoadType.ROLE:
            target_path = os.path.join(
                self.__path_mappings["ext_definitions"][LoadType.ROLE],
                ext_name,
            )
        elif ext_type == LoadType.COLLECTION:
            target_path = os.path.join(
                self.__path_mappings["ext_definitions"][LoadType.COLLECTION],
                ext_name,
            )
        else:
            raise ValueError("Invalid ext_type")
        return target_path

    def load_definition_ext(self, target_type, target_name, target_path):
        ld = self.create_load_file(target_type, target_name, target_path)
        use_cache = True
        output_dir = self.get_definition_path(ld.target_type, ld.target_name)
        if use_cache and os.path.exists(os.path.join(output_dir, "mappings.json")):
            if not self.silent:
                logger.debug("use cache from {}".format(output_dir))
            definitions, mappings = Parser.restore_definition_objects(output_dir)
        else:
            definitions, mappings = self._parser.run(load_data=ld)
            if self.do_save:
                if output_dir == "":
                    raise ValueError("Invalid output_dir")
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir, exist_ok=True)
                Parser.dump_definition_objects(output_dir, definitions, mappings)

        key = "{}-{}".format(target_type, target_name)
        self.ext_definitions[key] = {
            "definitions": definitions,
            "mappings": mappings,
        }
        return

    def _set_load_root(self, target_path=""):
        root_load_data = None
        if self.type in [LoadType.ROLE, LoadType.COLLECTION]:
            ext_type = self.type
            ext_name = self.name
            if target_path == "":
                target_path = self.get_source_path(ext_type, ext_name)
            root_load_data = self.create_load_file(ext_type, ext_name, target_path)
        elif self.type in [LoadType.PROJECT, LoadType.PLAYBOOK, LoadType.TASKFILE]:
            src_root = self.get_src_root()
            if target_path == "":
                target_path = os.path.join(src_root, escape_url(self.name))
            root_load_data = self.create_load_file(self.type, self.name, target_path)
        return root_load_data

    def get_source_path(self, ext_type, ext_name, is_ext_for_project=False):
        base_dir = ""
        if is_ext_for_project:
            base_dir = self.__path_mappings["dependencies"]
        else:
            if ext_type == LoadType.ROLE:
                base_dir = os.path.join(self.root_dir, "roles", "src")
            elif ext_type == LoadType.COLLECTION:
                base_dir = os.path.join(self.root_dir, "collections", "src")

        target_path = ""
        if ext_type == LoadType.ROLE:
            target_path = os.path.join(base_dir, ext_name)
        elif ext_type == LoadType.COLLECTION:
            parts = ext_name.split(".")
            target_path = os.path.join(
                base_dir,
                "ansible_collections",
                parts[0],
                parts[1],
            )
        else:
            raise ValueError("Invalid ext_type")
        return target_path

    def load_definitions_root(self, target_path=""):

        output_dir = self.__path_mappings["root_definitions"]
        root_load = self._set_load_root(target_path=target_path)

        definitions, mappings = self._parser.run(load_data=root_load, collection_name_of_project=self.collection_name)
        if self.do_save:
            if output_dir == "":
                raise ValueError("Invalid output_dir")
            if not os.path.exists(output_dir):
                os.makedirs(output_dir, exist_ok=True)
            Parser.dump_definition_objects(output_dir, definitions, mappings)

        self.root_definitions = {
            "definitions": definitions,
            "mappings": mappings,
        }

    def apply_spec_mutations(self):
        if not self.spec_mutations_from_previous_scan:
            return
        # overwrite the loaded object with the mutated object in spec mutations
        for type_name in self.root_definitions["definitions"]:
            obj_list = self.root_definitions["definitions"][type_name]
            for i, obj in enumerate(obj_list):
                key = obj.key
                if key in self.spec_mutations_from_previous_scan:
                    mutated_spec = self.spec_mutations_from_previous_scan[key].object
                    self.root_definitions["definitions"][type_name][i] = mutated_spec
        return

    def construct_trees(self, ram_client=None):
        trees, additional, extra_requirements, resolve_failures = tree(
            self.root_definitions, self.ext_definitions, ram_client, self.target_playbook_name, self.target_taskfile_name
        )

        # set annotation for spec mutations
        if self.spec_mutations_from_previous_scan:
            spec_mutations = self.spec_mutations_from_previous_scan
            for _tree in trees:
                for callobj in _tree.items:
                    if not isinstance(callobj, TaskCall):
                        continue
                    obj_key = callobj.spec.key
                    if obj_key in spec_mutations:
                        m = spec_mutations[obj_key]
                        rule_id = m.rule.rule_id
                        value = {
                            "rule_id": rule_id,
                            "changes": m.changes,
                        }
                        callobj.set_annotation(key="spec.mutations", value=value, rule_id=rule_id)

        self.trees = trees
        self.additional = additional
        self.extra_requirements = extra_requirements
        self.resolve_failures = resolve_failures

        if self.do_save:
            root_def_dir = self.__path_mappings["root_definitions"]
            tree_rel_file = os.path.join(root_def_dir, "tree.json")
            if tree_rel_file != "":
                lines = []
                for t_obj_list in self.trees:
                    lines.append(t_obj_list.to_one_line_json())
                open(tree_rel_file, "w").write("\n".join(lines))
                if not self.silent:
                    logger.info("  tree file saved")
        return

    def resolve_variables(self):
        taskcalls_in_trees = resolve(self.trees, self.additional)
        self.taskcalls_in_trees = taskcalls_in_trees

        for tree in self.trees:
            ctx = AnsibleRunContext.from_tree(tree)
            self.contexts.append(ctx)

        if self.do_save:
            root_def_dir = self.__path_mappings["root_definitions"]
            tasks_in_t_path = os.path.join(root_def_dir, "tasks_in_trees.json")
            tasks_in_t_lines = []
            for d in taskcalls_in_trees:
                line = jsonpickle.encode(d, make_refs=False)
                tasks_in_t_lines.append(line)

            open(tasks_in_t_path, "w").write("\n".join(tasks_in_t_lines))
        return

    def annotate(self):
        contexts = analyze(self.contexts)
        self.contexts = contexts

        if self.do_save:
            root_def_dir = self.__path_mappings["root_definitions"]
            contexts_a_path = os.path.join(root_def_dir, "contexts_with_analysis.json")
            conetxts_a_lines = []
            for d in contexts:
                line = jsonpickle.encode(d, make_refs=False)
                conetxts_a_lines.append(line)

            open(contexts_a_path, "w").write("\n".join(conetxts_a_lines))

        return

    def apply_rules(self):
        target_name = self.name
        if self.collection_name:
            target_name = self.collection_name
        if self.role_name:
            target_name = self.role_name
        data_report = detect(self.contexts, rules_dir=self.rules_dir, rules=self.rules)
        spec_mutations = data_report.get("spec_mutations", {})
        if spec_mutations:
            self.spec_mutations = spec_mutations
        metadata = {
            "type": self.type,
            "name": target_name,
            "version": self.version,
            "source": self.source_repository,
            "download_url": self.download_url,
            "hash": self.hash,
        }
        dependencies = self.loaded_dependency_dirs

        summary_txt = summarize_findings_data(
            metadata,
            dependencies,
            data_report,
            self.resolve_failures,
            self.extra_requirements,
        )
        self.findings = Findings(
            metadata=metadata,
            dependencies=dependencies,
            root_definitions=self.root_definitions,
            ext_definitions=self.ext_definitions,
            extra_requirements=self.extra_requirements,
            resolve_failures=self.resolve_failures,
            prm=self.prm,
            report=data_report,
            summary_txt=summary_txt,
            scan_time=datetime.datetime.utcnow().isoformat(),
        )
        self.result = data_report.get("ari_result", None)
        return

    def add_time_records(self, time_records: dict):
        if self.findings:
            self.findings.metadata["time_records"] = time_records
        return

    def count_definitions(self):
        dep_num = len(self.loaded_dependency_dirs)
        ext_counts = {}
        for _, _defs in self.ext_definitions.items():
            for key, val in _defs.get("definitions", {}).items():
                _current = ext_counts.get(key, 0)
                _current += len(val)
                ext_counts[key] = _current
        root_counts = {}
        for key, val in self.root_definitions.get("definitions", {}).items():
            _current = root_counts.get(key, 0)
            _current += len(val)
            root_counts[key] = _current
        return dep_num, ext_counts, root_counts

    def set_metadata(self, metadata: dict, dependencies: dict):
        self.target_path = self.make_target_path(self.type, self.name)
        self.version = metadata.get("version", "")
        self.hash = metadata.get("hash", "")
        self.download_url = metadata.get("download_url", "")
        self.loaded_dependency_dirs = dependencies

    def set_metadata_findings(self):
        target_name = self.name
        if self.collection_name:
            target_name = self.collection_name
        if self.role_name:
            target_name = self.role_name
        metadata = {
            "type": self.type,
            "name": target_name,
            "version": self.version,
            "source": self.source_repository,
            "download_url": self.download_url,
            "hash": self.hash,
        }
        dependencies = self.loaded_dependency_dirs
        self.findings = Findings(
            metadata=metadata,
            dependencies=dependencies,
        )

    def load_index(self):
        index_location = self.__path_mappings["index"]
        with open(index_location, "r") as f:
            self.index = json.load(f)


@dataclass
class ARIScanner(object):
    config: Config = None

    root_dir: str = ""
    rules_dir: str = ""
    rules: list = field(default_factory=list)

    ram_client: RAMClient = None
    read_ram: bool = True
    write_ram: bool = True

    skip_playbook_format_error: bool = (True,)
    skip_task_format_error: bool = (True,)

    use_ansible_doc: bool = True

    do_save: bool = False
    _parser: Parser = None

    show_all: bool = False
    pretty: bool = False
    silent: bool = False
    output_format: str = ""

    _current: SingleScan = None

    def __post_init__(self):
        if not self.config:
            self.config = config

        if not self.root_dir:
            self.root_dir = self.config.data_dir
        if not self.rules_dir:
            self.rules_dir = self.config.rules_dir
        if not self.rules:
            self.rules = self.config.rules
        self.ram_client = RAMClient(root_dir=self.root_dir)
        self._parser = Parser(
            do_save=self.do_save,
            use_ansible_doc=self.use_ansible_doc,
            skip_playbook_format_error=self.skip_playbook_format_error,
            skip_task_format_error=self.skip_task_format_error,
        )

        if not self.silent:
            logger.debug(f"config: {self.config}")

    def evaluate(
        self,
        type: str,
        name: str = "",
        path: str = "",
        collection_name: str = "",
        role_name: str = "",
        install_dependencies: bool = True,
        version: str = "",
        hash: str = "",
        target_path: str = "",
        dependency_dir: str = "",
        download_only: bool = False,
        use_src_cache: bool = False,
        source_repository: str = "",
        playbook_yaml: str = "",
        playbook_only: bool = False,
        taskfile_yaml: str = "",
        taskfile_only: bool = False,
        raw_yaml: str = "",
        out_dir: str = "",
        spec_mutations_from_previous_scan: dict = None,
    ):
        time_records = {}
        self.record_begin(time_records, "scandata_init")

        if not name and path:
            name = path

        if raw_yaml:
            if type == LoadType.PLAYBOOK:
                playbook_yaml = raw_yaml
            elif type == LoadType.TASKFILE:
                taskfile_yaml = raw_yaml

        if is_local_path(name):
            name = os.path.abspath(name)

        scandata = SingleScan(
            type=type,
            name=name,
            collection_name=collection_name,
            role_name=role_name,
            install_dependencies=install_dependencies,
            version=version,
            hash=hash,
            target_path=target_path,
            skip_playbook_format_error=self.skip_playbook_format_error,
            skip_task_format_error=self.skip_task_format_error,
            dependency_dir=dependency_dir,
            use_src_cache=use_src_cache,
            source_repository=source_repository,
            playbook_yaml=playbook_yaml,
            playbook_only=playbook_only,
            taskfile_yaml=taskfile_yaml,
            taskfile_only=taskfile_only,
            out_dir=out_dir,
            root_dir=self.root_dir,
            rules_dir=self.rules_dir,
            rules=self.rules,
            spec_mutations_from_previous_scan=spec_mutations_from_previous_scan,
            use_ansible_doc=self.use_ansible_doc,
            do_save=self.do_save,
            silent=self.silent,
            _parser=self._parser,
        )
        self._current = scandata
        self.record_end(time_records, "scandata_init")

        self.record_begin(time_records, "metadata_load")
        metdata_loaded = False
        if self.read_ram:
            loaded, metadata, dependencies = self.load_metadata_from_ram(scandata.type, scandata.name, scandata.version)
            logger.debug(f"metadata loaded: {loaded}")
            if loaded:
                scandata.set_metadata(metadata, dependencies)
                metdata_loaded = True
                if not self.silent:
                    logger.debug(f'Use metadata for "{scandata.name}" in RAM DB')

        if scandata.install_dependencies and not metdata_loaded:
            print(f"start preparing {scandata.type} {scandata.name}")
            scandata._prepare_dependencies()
            print(f"finished preparing {scandata.type} {scandata.name}")

        if download_only:
            scandata.set_metadata_findings()
            self.register_findings_to_ram(scandata.findings)
            return None
        self.record_end(time_records, "metadata_load")

        ext_list = []
        ext_list.extend(
            [
                (
                    d.get("metadata", {}).get("type", ""),
                    d.get("metadata", {}).get("name", ""),
                    d.get("metadata", {}).get("version", ""),
                    d.get("metadata", {}).get("hash", ""),
                    d.get("dir"),
                )
                for d in scandata.loaded_dependency_dirs
            ]
        )
        ext_count = len(ext_list)

        # PRM Finder
        self.record_begin(time_records, "prm_load")
        # playbooks, roles, modules = find_playbook_role_module(scandata.target_path, self.use_ansible_doc)
        # scandata.prm["playbooks"] = playbooks
        # scandata.prm["roles"] = roles
        # scandata.prm["modules"] = modules
        self.record_end(time_records, "prm_load")

        # Start ARI Scanner main flow
        self.record_begin(time_records, "dependency_load")
        for i, (ext_type, ext_name, ext_ver, ext_hash, ext_path) in enumerate(ext_list):
            if not self.silent:
                if i == 0:
                    logger.info("start loading {} {}(s)".format(ext_count, ext_type))
                logger.info("[{}/{}] {} {}".format(i + 1, ext_count, ext_type, ext_name))

            # avoid infinite loop
            is_root = False
            if scandata.type == ext_type and scandata.name == ext_name:
                is_root = True

            if not is_root:
                # scan dependencies and save findings to ARI RAM
                dep_scanner = ARIScanner(
                    root_dir=self.root_dir,
                    rules_dir=self.rules_dir,
                    rules=self.rules,
                    ram_client=self.ram_client,
                    read_ram=self.read_ram,
                    write_ram=self.write_ram,
                    use_ansible_doc=self.use_ansible_doc,
                    do_save=self.do_save,
                    silent=True,
                )
                # use prepared dep dirs
                dep_scanner.evaluate(
                    type=ext_type,
                    name=ext_name,
                    version=ext_ver,
                    hash=ext_hash,
                    target_path=ext_path,
                    dependency_dir=scandata.dependency_dir,
                    source_repository=scandata.source_repository,
                )

                if self.read_ram:
                    # searching findings from ARI RAM and use them if found
                    loaded, ext_defs = self.load_definitions_from_ram(ext_type, ext_name, ext_ver, ext_hash)
                    if loaded:
                        key = "{}-{}".format(ext_type, ext_name)
                        scandata.ext_definitions[key] = ext_defs
                        if not self.silent:
                            logger.debug(f'Use spec data for "{ext_name}" in RAM DB')
                        continue

                # load the src directory
                scandata.load_definition_ext(ext_type, ext_name, ext_path)
        self.record_end(time_records, "dependency_load")

        if not self.silent:
            logger.debug("load_definition_ext() done")

        loaded = False
        self.record_begin(time_records, "target_load")
        if self.read_ram and scandata.type not in [LoadType.PLAYBOOK, LoadType.TASKFILE]:
            loaded, root_defs = self.load_definitions_from_ram(scandata.type, scandata.name, scandata.version, scandata.hash, allow_unresolved=True)
            logger.debug(f"spec data loaded: {loaded}")
            if loaded:
                scandata.root_definitions = root_defs
                if not self.silent:
                    logger.info("Use spec data in RAM DB")
        self.record_end(time_records, "target_load")

        if not loaded:
            scandata.load_definitions_root(target_path=scandata.target_path)

        if not self.silent:
            logger.debug("load_definitions_root() done")
            playbooks_num = len(scandata.root_definitions["definitions"]["playbooks"])
            roles_num = len(scandata.root_definitions["definitions"]["roles"])
            taskfiles_num = len(scandata.root_definitions["definitions"]["taskfiles"])
            tasks_num = len(scandata.root_definitions["definitions"]["tasks"])
            modules_num = len(scandata.root_definitions["definitions"]["modules"])
            logger.debug(f"playbooks: {playbooks_num}, roles: {roles_num}, taskfiles: {taskfiles_num}, tasks: {tasks_num}, modules: {modules_num}")

        self.record_begin(time_records, "apply_spec_rules")
        scandata.apply_spec_mutations()
        self.record_end(time_records, "apply_spec_rules")
        if not self.silent:
            logger.debug("apply_spec_rules() done")

        _ram_client = None
        if self.read_ram:
            _ram_client = self.ram_client

        self.record_begin(time_records, "tree_construction")
        scandata.construct_trees(_ram_client)
        self.record_end(time_records, "tree_construction")
        if not self.silent:
            logger.debug("construct_trees() done")

        self.record_begin(time_records, "variable_resolution")
        scandata.resolve_variables()
        self.record_end(time_records, "variable_resolution")
        if not self.silent:
            logger.debug("resolve_variables() done")

        self.record_begin(time_records, "module_annotators")
        scandata.annotate()
        if not self.silent:
            logger.debug("annotate() done")
        self.record_end(time_records, "module_annotators")

        self.record_begin(time_records, "apply_rules")
        scandata.apply_rules()
        self.record_end(time_records, "apply_rules")
        if not self.silent:
            logger.debug("apply_rules() done")

        scandata.add_time_records(time_records=time_records)

        dep_num, ext_counts, root_counts = scandata.count_definitions()
        if not self.silent:
            print("# of dependencies:", dep_num)
            # print("ext definitions:", ext_counts)
            # print("root definitions:", root_counts)

        # save RAM data
        if self.write_ram:
            self.register_findings_to_ram(scandata.findings)
            self.register_module_index_to_ram(scandata.findings)

        if scandata.out_dir is not None and scandata.out_dir != "":
            self.save_rule_result(scandata.findings, scandata.out_dir)
            if not self.silent:
                print("The rule result is saved at {}".format(scandata.out_dir))

        if not self.silent:
            summary = summarize_findings(scandata.findings, self.show_all)
            print(summary)

        if self.pretty:
            data_str = ""
            data = json.loads(jsonpickle.encode(scandata.findings.simple(), make_refs=False))
            if self.output_format.lower() == "json":
                data_str = json.dumps(data, indent=2)
            elif self.output_format.lower() == "yaml":
                data_str = yaml.safe_dump(data)
            print(data_str)

        if scandata.spec_mutations:
            trigger_rescan = False
            _previous = spec_mutations_from_previous_scan
            if _previous and equal(scandata.spec_mutations, _previous):
                if not self.silent:
                    logger.warning("Spec mutation loop has been detected! " "Exitting the scan here but the result may be incomplete.")
            else:
                trigger_rescan = True

            if trigger_rescan:
                if not self.silent:
                    print("Spec mutations are found. Triggering ARI scan again...")
                return self.evaluate(
                    type=type,
                    name=name,
                    path=path,
                    collection_name=collection_name,
                    role_name=role_name,
                    install_dependencies=install_dependencies,
                    version=version,
                    hash=hash,
                    target_path=target_path,
                    dependency_dir=dependency_dir,
                    download_only=download_only,
                    use_src_cache=use_src_cache,
                    source_repository=source_repository,
                    playbook_yaml=playbook_yaml,
                    playbook_only=playbook_only,
                    taskfile_yaml=taskfile_yaml,
                    taskfile_only=taskfile_only,
                    raw_yaml=raw_yaml,
                    out_dir=out_dir,
                    spec_mutations_from_previous_scan=scandata.spec_mutations,
                )

        return scandata.findings.report.get("ari_result", None)

    def load_metadata_from_ram(self, type, name, version):
        loaded, metadata, dependencies = self.ram_client.load_metadata_from_findings(type, name, version)
        return loaded, metadata, dependencies

    def load_definitions_from_ram(self, type, name, version, hash, allow_unresolved=False):
        loaded, definitions, mappings = self.ram_client.load_definitions_from_findings(type, name, version, hash, allow_unresolved)
        definitions_dict = {}
        if loaded:
            definitions_dict = {
                "definitions": definitions,
                "mappings": mappings,
            }
        return loaded, definitions_dict

    def register_findings_to_ram(self, findings: Findings):
        self.ram_client.register(findings)

    def register_module_index_to_ram(self, findings: Findings):
        self.ram_client.module_index_register(findings)

    def save_findings(self, findings: Findings, out_dir: str):
        self.ram_client.save_findings(findings, out_dir)

    def save_rule_result(self, findings: Findings, out_dir: str):
        if out_dir == "":
            raise ValueError("output dir must be a non-empty value")

        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        findings.save_rule_result(fpath=os.path.join(out_dir, "rule_result.json"))

    def get_last_scandata(self):
        return self._current

    def save_error(self, error: str, out_dir: str = ""):
        if out_dir == "":
            type = self._current.type
            name = self._current.name
            version = self._current.version
            hash = self._current.hash
            out_dir = self.ram_client.make_findings_dir_path(type, name, version, hash)
        self.ram_client.save_error(error, out_dir)

    def record_begin(self, time_records: dict, record_name: str):
        time_records[record_name] = {}
        time_records[record_name]["begin"] = datetime.datetime.utcnow().isoformat()

    def record_end(self, time_records: dict, record_name: str):
        end = datetime.datetime.utcnow()
        time_records[record_name]["end"] = end.isoformat()
        begin = datetime.datetime.fromisoformat(time_records[record_name]["begin"])
        elapsed = (end - begin).total_seconds()
        time_records[record_name]["elapsed"] = elapsed


def tree(root_definitions, ext_definitions, ram_client=None, target_playbook_path=None, target_taskfile_path=None):
    tl = TreeLoader(root_definitions, ext_definitions, ram_client, target_playbook_path, target_taskfile_path)
    trees, additional = tl.run()
    if trees is None:
        raise ValueError("failed to get trees")
    # if node_objects is None:
    #     raise ValueError("failed to get node_objects")
    return (
        trees,
        additional,
        tl.extra_requirements,
        tl.resolve_failures,
    )


def resolve(trees, additional):
    taskcalls_in_trees = []
    for i, tree in enumerate(trees):
        if not isinstance(tree, ObjectList):
            continue
        if len(tree.items) == 0:
            continue
        root_key = tree.items[0].spec.key
        logger.debug("[{}/{}] {}".format(i + 1, len(trees), root_key))
        taskcalls = resolve_variables(tree, additional)
        d = TaskCallsInTree(
            root_key=root_key,
            taskcalls=taskcalls,
        )
        taskcalls_in_trees.append(d)
    return taskcalls_in_trees


if __name__ == "__main__":
    __target_type = sys.argv[1]
    __target_name = sys.argv[2]
    __dependency_dir = ""
    if len(sys.argv) >= 4:
        __dependency_dir = sys.argv[3]
    c = ARIScanner(
        root_dir=config.data_dir,
    )
    c.evaluate(
        type=__target_type,
        name=__target_name,
        dependency_dir=__dependency_dir,
    )



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    Rule,
    Severity,
    RuleTag as Tag,
    RuleResult,
)


@dataclass
class UnusedOverrideRule(Rule):
    rule_id: str = "R203"
    description: str = "A variable is not successfully re-defined because of low precedence"
    enabled: bool = True
    name: str = "UnusedOverride"
    version: str = "v0.0.1"
    severity: Severity = Severity.VERY_LOW
    tags: tuple = Tag.VARIABLE

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        verdict = False
        detail = {"variables": []}
        if task.spec.defined_vars:
            for v in task.spec.defined_vars:
                all_definitions = task.variable_set.get(v, [])
                if len(all_definitions) > 1:
                    prev_prec = all_definitions[-2].type
                    new_prec = all_definitions[-1].type
                    if new_prec < prev_prec:
                        detail["variables"].append(
                            {
                                "name": v,
                                "prev_precedence": prev_prec,
                                "new_precedence": new_prec,
                            }
                        )
                        verdict = True

        return RuleResult(verdict=verdict, detail=detail, file=task.file_info(), rule=self.get_metadata())



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass, field
from typing import List, Union
from collections.abc import Callable
from tabulate import tabulate
from ruamel.yaml.scalarstring import DoubleQuotedScalarString

from copy import deepcopy
import json
import jsonpickle
import ansible_risk_insight.yaml as ariyaml
from ansible.module_utils.parsing.convert_bool import boolean
from .keyutil import (
    set_collection_key,
    set_module_key,
    set_play_key,
    set_playbook_key,
    set_repository_key,
    set_role_key,
    set_task_key,
    set_taskfile_key,
    set_call_object_key,
    get_obj_info_by_key,
)
from .utils import (
    equal,
    recursive_copy_dict,
)


class PlaybookFormatError(Exception):
    pass


class TaskFormatError(Exception):
    pass


class JSONSerializable(object):
    def dump(self):
        return self.to_json()

    def to_json(self):
        return jsonpickle.encode(self, make_refs=False)

    @classmethod
    def from_json(cls, json_str):
        instance = cls()
        loaded = jsonpickle.decode(json_str)
        instance.__dict__.update(loaded.__dict__)
        return instance


class Resolvable(object):
    def resolve(self, resolver):
        if not hasattr(resolver, "apply"):
            raise ValueError("this resolver does not have apply() method")
        if not callable(resolver.apply):
            raise ValueError("resolver.apply is not callable")

        # apply resolver for this instance
        resolver.apply(self)

        # call resolve() for children recursively
        targets = self.resolver_targets
        if targets is None:
            return
        for t in targets:
            if isinstance(t, str):
                continue
            t.resolve(resolver)

        # apply resolver again here
        # because some attributes was not set at first
        resolver.apply(self)
        return

    @property
    def resolver_targets(self):
        raise NotImplementedError


class LoadType:
    PROJECT = "project"
    COLLECTION = "collection"
    ROLE = "role"
    PLAYBOOK = "playbook"
    TASKFILE = "taskfile"
    UNKNOWN = "unknown"


@dataclass
class Load(JSONSerializable):
    target_name: str = ""
    target_type: str = ""
    path: str = ""
    loader_version: str = ""
    playbook_yaml: str = ""
    playbook_only: bool = False
    taskfile_yaml: str = ""
    taskfile_only: bool = False
    timestamp: str = ""

    # the following variables are list of paths; not object
    roles: list = field(default_factory=list)
    playbooks: list = field(default_factory=list)
    taskfiles: list = field(default_factory=list)
    modules: list = field(default_factory=list)


@dataclass
class Object(JSONSerializable):
    type: str = ""
    key: str = ""


@dataclass
class ObjectList(JSONSerializable):
    items: list = field(default_factory=list)
    _dict: dict = field(default_factory=dict)

    def dump(self, fpath=""):
        return self.to_json(fpath=fpath)

    def to_json(self, fpath=""):
        lines = [jsonpickle.encode(obj, make_refs=False) for obj in self.items]
        json_str = "\n".join(lines)
        if fpath != "":
            open(fpath, "w").write(json_str)
        return json_str

    def to_one_line_json(self):
        return jsonpickle.encode(self.items, make_refs=False)

    @classmethod
    def from_json(cls, json_str="", fpath=""):
        instance = cls()
        if fpath != "":
            json_str = open(fpath, "r").read()
        lines = json_str.splitlines()
        items = [jsonpickle.decode(obj_str) for obj_str in lines]
        instance.items = items
        instance._update_dict()
        return instance

    def add(self, obj, update_dict=True):
        self.items.append(obj)
        if update_dict:
            self._add_dict_item(obj)
        return

    def merge(self, obj_list):
        if not isinstance(obj_list, ObjectList):
            raise ValueError("obj_list must be an instance of ObjectList, but got {}".format(type(obj_list).__name__))
        self.items.extend(obj_list.items)
        self._update_dict()
        return

    def find_by_attr(self, key, val):
        found = [obj for obj in self.items if obj.__dict__.get(key, None) == val]
        return found

    def find_by_type(self, type):
        return [obj for obj in self.items if hasattr(obj, "type") and obj.type == type]

    def find_by_key(self, key):
        return self._dict.get(key, None)

    def contains(self, key="", obj=None):
        if obj is not None:
            key = obj.key
        return self.find_by_key(key) is not None

    def update_dict(self):
        self._update_dict()

    def _update_dict(self):
        for obj in self.items:
            self._dict[obj.key] = obj
        return

    def _add_dict_item(self, obj):
        self._dict[obj.key] = obj

    @property
    def resolver_targets(self):
        return self.items


@dataclass
class CallObject(JSONSerializable):
    type: str = ""
    key: str = ""
    called_from: str = ""
    spec: Object = field(default_factory=Object)
    depth: int = -1
    node_id: str = ""

    @classmethod
    def from_spec(cls, spec, caller, index):
        instance = cls()
        instance.spec = spec
        caller_key = "None"
        depth = 0
        node_id = "0"
        if caller:
            instance.called_from = caller.key
            caller_key = caller.key
            depth = caller.depth + 1
            index_str = "0"
            if index >= 0:
                index_str = str(index)
            node_id = caller.node_id + "." + index_str
        instance.depth = depth
        instance.node_id = node_id
        instance.key = set_call_object_key(cls.__name__, spec.key, caller_key)
        return instance


class RunTargetType:
    Playbook = "playbookcall"
    Play = "playcall"
    Role = "rolecall"
    TaskFile = "taskfilecall"
    Task = "taskcall"


@dataclass
class RunTarget(object):
    type: str = ""

    def file_info(self):
        file = self.spec.defined_in
        lines = None
        return file, lines


@dataclass
class RunTargetList(object):
    items: List[RunTarget] = field(default_factory=list)

    _i: int = 0

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return self

    def __next__(self):
        if self._i == len(self.items):
            self._i = 0
            raise StopIteration()
        item = self.items[self._i]
        self._i += 1
        return item

    def __getitem__(self, i):
        return self.items[i]


@dataclass
class ModuleArgument(object):
    name: str = ""
    type: str = None
    default: any = None
    required: bool = False
    description: str = ""
    choices: list = field(default_factory=list)
    aliases: list = field(default_factory=list)

    def available_keys(self):
        keys = [self.name]
        if self.aliases:
            keys.extend(self.aliases)
        return keys


@dataclass
class Module(Object, Resolvable):
    type: str = "module"
    name: str = ""
    fqcn: str = ""
    key: str = ""
    local_key: str = ""
    collection: str = ""
    role: str = ""
    documentation: str = ""
    examples: str = ""
    arguments: list = field(default_factory=list)
    defined_in: str = ""
    builtin: bool = False
    used_in: list = field(default_factory=list)  # resolved later

    annotations: dict = field(default_factory=dict)

    def set_key(self):
        set_module_key(self)

    def children_to_key(self):
        return self

    @property
    def resolver_targets(self):
        return None


@dataclass
class ModuleCall(CallObject, Resolvable):
    type: str = "modulecall"


@dataclass
class ModuleMetadata(object):
    fqcn: str = ""
    # arguments: list = field(default_factory=list)
    type: str = ""
    name: str = ""
    version: str = ""
    hash: str = ""

    @staticmethod
    def from_module(m: Module, metadata: dict):
        mm = ModuleMetadata()
        for key in mm.__dict__:
            if hasattr(m, key):
                val = getattr(m, key, None)
                setattr(mm, key, val)

        mm.type = metadata.get("type", "")
        mm.name = metadata.get("name", "")
        mm.version = metadata.get("version", "")
        mm.hash = metadata.get("hash", "")
        return mm


@dataclass
class Collection(Object, Resolvable):
    type: str = "collection"
    name: str = ""
    path: str = ""
    key: str = ""
    local_key: str = ""
    metadata: dict = field(default_factory=dict)
    files: dict = field(default_factory=dict)
    playbooks: list = field(default_factory=list)
    taskfiles: list = field(default_factory=list)
    roles: list = field(default_factory=list)
    modules: list = field(default_factory=list)
    dependency: dict = field(default_factory=dict)
    requirements: dict = field(default_factory=dict)

    annotations: dict = field(default_factory=dict)

    variables: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)

    def set_key(self):
        set_collection_key(self)

    def children_to_key(self):
        module_keys = [m.key if isinstance(m, Module) else m for m in self.modules]
        self.modules = sorted(module_keys)

        playbook_keys = [p.key if isinstance(p, Playbook) else p for p in self.playbooks]
        self.playbooks = sorted(playbook_keys)

        role_keys = [r.key if isinstance(r, Role) else r for r in self.roles]
        self.roles = sorted(role_keys)

        taskfile_keys = [tf.key if isinstance(tf, TaskFile) else tf for tf in self.taskfiles]
        self.taskfiles = sorted(taskfile_keys)
        return self

    @property
    def resolver_targets(self):
        return self.playbooks + self.taskfiles + self.roles + self.modules


@dataclass
class CollectionCall(CallObject, Resolvable):
    type: str = "collectioncall"


@dataclass
class TaskCallsInTree(JSONSerializable):
    root_key: str = ""
    taskcalls: list = field(default_factory=list)


@dataclass
class VariablePrecedence(object):
    name: str = ""
    order: int = -1

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __eq__(self, __o: object) -> bool:
        return self.order == __o.order

    def __ne__(self, __o: object) -> bool:
        return not self.__eq__(__o)

    def __lt__(self, __o: object):
        return self.order < __o.order

    def __le__(self, __o: object):
        return self.__lt__(__o) or self.__eq__(__o)

    def __gt__(self, __o: object):
        return not self.__le__(__o)

    def __ge__(self, __o: object):
        return not self.__lt__(__o)


class VariableType(object):
    # When resolving variables, sometimes find unknown variables (e.g. undefined variable)
    # so we consider it as one type of variable
    Unknown = VariablePrecedence("unknown", -100)
    # Variable Precedence
    # https://docs.ansible.com/ansible/latest/playbook_guide
    #     /playbooks_variables.html#understanding-variable-precedence
    CommandLineValues = VariablePrecedence("command_line_values", 1)
    RoleDefaults = VariablePrecedence("role_defaults", 2)
    InventoryFileOrScriptGroupVars = VariablePrecedence("inventory_file_or_script_group_vars", 3)
    InventoryGroupVarsAll = VariablePrecedence("inventory_group_vars_all", 4)
    PlaybookGroupVarsAll = VariablePrecedence("playbook_group_vars_all", 5)
    InventoryGroupVarsAny = VariablePrecedence("inventory_group_vars_any", 6)
    PlaybookGroupVarsAny = VariablePrecedence("playbook_group_vars_any", 7)
    InventoryFileOrScriptHostVars = VariablePrecedence("inventory_file_or_script_host_vars", 8)
    InventoryHostVarsAny = VariablePrecedence("inventory_host_vars_any", 9)
    PlaybookHostVarsAny = VariablePrecedence("playbook_host_vars_any", 10)
    HostFacts = VariablePrecedence("host_facts", 11)
    PlayVars = VariablePrecedence("play_vars", 12)
    PlayVarsPrompt = VariablePrecedence("play_vars_prompt", 13)
    PlayVarsFiles = VariablePrecedence("play_vars_files", 14)
    RoleVars = VariablePrecedence("role_vars", 15)
    BlockVars = VariablePrecedence("block_vars", 16)
    TaskVars = VariablePrecedence("task_vars", 17)
    IncludeVars = VariablePrecedence("include_vars", 18)
    # we deal with set_facts and registered_vars separately
    # because the expression in a fact will be evaluated everytime it is used
    SetFacts = VariablePrecedence("set_facts", 19)
    RegisteredVars = VariablePrecedence("registered_vars", 20)
    RoleParams = VariablePrecedence("role_params", 21)
    IncludeParams = VariablePrecedence("include_params", 22)
    ExtraVars = VariablePrecedence("extra_vars", 23)
    # vars defined in `loop` cannot be overridden by the vars above
    # so we put this as a highest precedence var type
    LoopVars = VariablePrecedence("loop_vars", 24)


immutable_var_types = [VariableType.LoopVars]


@dataclass
class Variable(object):
    name: str = ""
    value: any = None
    type: VariableType = None
    elements: list = field(default_factory=list)
    setter: any = None
    used_in: any = None

    @property
    def is_mutable(self):
        return self.type not in immutable_var_types


@dataclass
class VariableDict(object):
    _dict: dict = field(default_factory=dict)

    @staticmethod
    def print_table(data: dict):
        d = VariableDict(_dict=data)
        table = []
        type_labels = []
        found_type_label_names = []
        for v_list in d._dict.values():
            for v in v_list:
                if v.type.name in found_type_label_names:
                    continue
                type_labels.append(v.type)
                found_type_label_names.append(v.type.name)
        type_labels = sorted(type_labels, key=lambda x: x.order, reverse=True)

        for v_name in d._dict:
            v_list = d._dict[v_name]
            row = {"NAME": v_name}
            for t in type_labels:
                value = "-"
                for v in v_list:
                    if v.type != t:
                        continue
                    value = v.value
                    if isinstance(value, str) and value == "":
                        value = '""'
                type_label = t.name.upper()
                row[type_label] = value
            table.append(row)
        return tabulate(table, headers="keys")


class ArgumentsType(object):
    SIMPLE = "simple"
    LIST = "list"
    DICT = "dict"


@dataclass
class Arguments(object):
    type: ArgumentsType = ""
    raw: any = None
    vars: List[Variable] = field(default_factory=list)
    resolved: bool = False
    templated: any = None
    is_mutable: bool = False

    def get(self, key: str = ""):
        sub_raw = None
        sub_templated = None
        if key == "":
            sub_raw = self.raw
            sub_templated = self.templated
        else:
            if isinstance(self.raw, dict):
                sub_raw = self.raw.get(key, None)
                if self.templated:
                    sub_templated = self.templated[0].get(key, None)
            else:
                sub_raw = self.raw
                sub_templated = self.templated
        if not sub_raw:
            return None

        _vars = []
        sub_type = ArgumentsType.SIMPLE
        if isinstance(sub_raw, str):
            for v in self.vars:
                if v.name in sub_raw:
                    _vars.append(v)
        elif isinstance(sub_raw, list):
            sub_type = ArgumentsType.LIST
        elif isinstance(sub_raw, dict):
            sub_type = ArgumentsType.DICT
        is_mutable = False
        for v in _vars:
            if v.is_mutable:
                is_mutable = True
                break

        return Arguments(
            type=sub_type,
            raw=sub_raw,
            vars=_vars,
            resolved=self.resolved,
            templated=sub_templated,
            is_mutable=is_mutable,
        )


class LocationType:
    FILE = "file"
    DIR = "dir"
    URL = "url"


@dataclass
class Location(object):
    type: str = ""
    value: str = ""
    vars: List[Variable] = field(default_factory=list)

    _args: Arguments = None

    def __post_init__(self):
        if self._args:
            self.value = self._args.raw
            self.vars = self._args.vars

    @property
    def is_mutable(self):
        return len(self.vars) > 0

    @property
    def is_empty(self):
        return not self.type and not self.value

    def is_inside(self, loc):
        if not isinstance(loc, Location):
            raise ValueError(f"is_inside() expect Location but given {type(loc)}")
        return loc.contains(self)

    def contains(self, target, any=False, all=True):
        if isinstance(target, list):
            if any:
                return self.contains_any(target_list=target)
            elif all:
                return self.contains_all(target_list=target)
            else:
                raise ValueError('contains() must be run in either "any" or "all" mode')

        else:
            if not isinstance(target, Location):
                raise ValueError(f"contains() expect Location or list of Location, but given {type(target)}")

        my_path = self.value
        target_path = target.value
        if target_path.startswith(my_path):
            return True
        return False

    def contains_any(self, target_list):
        for target in target_list:
            if self.contains(target):
                return True
        return False

    def contains_all(self, target_list):
        count = 0
        for target in target_list:
            if self.contains(target):
                count += 1
        if count == len(target_list):
            return True
        return False


class AnnotationDetail(object):
    pass


@dataclass
class NetworkTransferDetail(AnnotationDetail):
    src: Location = None
    dest: Location = None
    is_mutable_src: bool = False
    is_mutable_dest: bool = False

    _src_arg: Arguments = None
    _dest_arg: Arguments = None

    def __post_init__(self):
        if self._src_arg:
            self.src = Location(_args=self._src_arg)
            if self._src_arg.is_mutable:
                self.is_mutable_src = True

        if self._dest_arg:
            self.dest = Location(_args=self._dest_arg)
            if self._dest_arg.is_mutable:
                self.is_mutable_dest = True


@dataclass
class InboundTransferDetail(NetworkTransferDetail):
    def __post_init__(self):
        super().__post_init__()


@dataclass
class OutboundTransferDetail(NetworkTransferDetail):
    def __post_init__(self):
        super().__post_init__()


@dataclass
class PackageInstallDetail(AnnotationDetail):
    pkg: any = ""
    version: any = ""
    is_mutable_pkg: bool = False
    disable_validate_certs: bool = False
    allow_downgrade: bool = False

    _pkg_arg: Arguments = None
    _version_arg: Arguments = None
    _allow_downgrade_arg: Arguments = None
    _validate_certs_arg: Arguments = None

    def __post_init__(self):
        if self._pkg_arg:
            self.pkg = self._pkg_arg.raw
            if self._pkg_arg.is_mutable:
                self.is_mutable_pkg = True
        if self._version_arg:
            self.version = self._version_arg.vars
        if self._allow_downgrade_arg:
            if _convert_to_bool(self._allow_downgrade_arg.raw):
                self.allow_downgrade = True
        if self._validate_certs_arg:
            if not _convert_to_bool(self._validate_certs_arg.raw):
                self.disable_validate_certs = True


@dataclass
class KeyConfigChangeDetail(AnnotationDetail):
    is_deletion: bool = False
    is_mutable_key: bool = False
    key: str = ""

    _key_arg: Arguments = None
    _state_arg: Arguments = None

    def __post_init__(self):
        if self._key_arg:
            self.key = self._key_arg.vars
            if self._key_arg.is_mutable:
                self.is_mutable_key = True
        if self._state_arg.raw == "absent":
            self.is_deletion = True


@dataclass
class FileChangeDetail(AnnotationDetail):
    path: Location = None
    src: Location = None
    is_mutable_path: bool = False
    is_mutable_src: bool = False
    is_unsafe_write: bool = False
    is_deletion: bool = False
    is_insecure_permissions: bool = False

    _path_arg: Arguments = None
    _src_arg: Arguments = None
    _mode_arg: Arguments = None
    _state_arg: Arguments = None
    _unsafe_write_arg: Arguments = None

    def __post_init__(self):
        if self._mode_arg and self._mode_arg.raw in ["1777", "0777"]:
            self.is_insecure_permissions = True
        if self._state_arg and self._state_arg.raw == "absent":
            self.is_deletion = True
        if self._path_arg:
            self.path = Location(_args=self._path_arg)
            if self._path_arg.is_mutable:
                self.is_mutable_path = True
        if self._src_arg:
            self.src = Location(_args=self._src_arg)
            if self._src_arg.is_mutable:
                self.is_mutable_src = True
        if self._unsafe_write_arg:
            if _convert_to_bool(self._unsafe_write_arg.raw):
                self.is_unsafe_write = True


execution_programs: list = ["sh", "bash", "zsh", "fish", "ash", "python*", "java*", "node*"]
non_execution_programs: list = ["tar", "gunzip", "unzip", "mv", "cp"]


@dataclass
class CommandExecDetail(AnnotationDetail):
    command: Arguments = None
    exec_files: List[Location] = field(default_factory=list)

    def __post_init__(self):
        self.exec_files = self.extract_exec_files()

    def extract_exec_files(self):
        cmd_str = self.command.raw
        if isinstance(cmd_str, list):
            cmd_str = " ".join(cmd_str)
        if isinstance(cmd_str, dict):
            cmd_str = cmd_str.get("cmd", "")
        lines = cmd_str.splitlines()
        exec_files = []
        for line in lines:
            parts = []
            is_in_variable = False
            concat_p = ""
            for p in line.split(" "):
                if "{{" in p and "}}" not in p:
                    is_in_variable = True
                if "}}" in p:
                    is_in_variable = False
                concat_p += " " + p if concat_p != "" else p
                if not is_in_variable:
                    parts.append(concat_p)
                    concat_p = ""
            found_program = None
            for i, p in enumerate(parts):
                if i == 0:
                    program = p if "/" not in p else p.split("/")[-1]
                    # filter out some specific non-exec patterns
                    if program in non_execution_programs:
                        break
                    # if the command string is like "python {{ python_script_path }}",
                    # {{ python_script_path }} is the exec file instead of "python"
                    if program in execution_programs:
                        continue
                    # for the case that the program name is like "python-3.6"
                    for exec_p in execution_programs:
                        if exec_p[-1] == "*":
                            if program.startswith(exec_p[:-1]):
                                continue
                if p.startswith("-"):
                    continue
                if found_program is None:
                    found_program = p
                    break
            if found_program:
                exec_file_name = found_program
                related_vars = [v for v in self.command.vars if v.name in exec_file_name]
                location_type = LocationType.FILE
                exec_file = Location(
                    type=location_type,
                    value=exec_file_name,
                    vars=related_vars,
                )
                exec_files.append(exec_file)
        return exec_files


def _convert_to_bool(a: any):
    if type(a) is bool:
        return bool(a)
    if type(a) is str:
        if a == "true" or a == "True" or a == "yes":
            return True
        else:
            return False
    return None


@dataclass
class Annotation(JSONSerializable):
    key: str = ""
    value: any = None

    rule_id: str = ""

    # TODO: avoid Annotation variants and remove `type`
    type: str = ""


@dataclass
class VariableAnnotation(Annotation):
    type: str = "variable_annotation"
    option_value: Arguments = field(default_factory=Arguments)


class RiskType:
    pass


class DefaultRiskType(RiskType):
    NONE = ""
    CMD_EXEC = "cmd_exec"
    INBOUND = "inbound_transfer"
    OUTBOUND = "outbound_transfer"
    FILE_CHANGE = "file_change"
    SYSTEM_CHANGE = "system_change"
    NETWORK_CHANGE = "network_change"
    CONFIG_CHANGE = "config_change"
    PACKAGE_INSTALL = "package_install"
    PRIVILEGE_ESCALATION = "privilege_escalation"


@dataclass
class RiskAnnotation(Annotation, NetworkTransferDetail, CommandExecDetail):
    type: str = "risk_annotation"
    risk_type: RiskType = ""

    @classmethod
    def init(cls, risk_type: RiskType, detail: AnnotationDetail):
        anno = cls()
        anno.risk_type = risk_type
        for attr_name in detail.__annotations__:
            val = getattr(detail, attr_name)
            setattr(anno, attr_name, val)
        return anno

    def equal_to(self, anno):
        if self.type != anno.type:
            return False
        if self.risk_type != anno.risk_type:
            return False
        self_dict = self.__dict__
        anno_dict = anno.__dict__
        if not equal(self_dict, anno_dict):
            return False
        return True


@dataclass
class FindCondition(object):
    def check(self, anno: RiskAnnotation):
        raise NotImplementedError


@dataclass
class AnnotationCondition(object):
    type: RiskType = ""
    attr_conditions: list = field(default_factory=list)

    def risk_type(self, risk_type: RiskType):
        self.type = risk_type
        return self

    def attr(self, key: str, val: any):
        self.attr_conditions.append((key, val))
        return self


@dataclass
class AttributeCondition(FindCondition):
    attr: str = None
    result: any = None

    def check(self, anno: RiskAnnotation):
        if self.attr:
            if hasattr(anno.detail, self.attr):
                anno_value = getattr(anno.detail, self.attr)
                if anno_value == self.result:
                    return True
                if self.result is None:
                    if isinstance(anno_value, bool) and anno_value:
                        return True
        return False


@dataclass
class FunctionCondition(FindCondition):
    func: Callable = None
    args: List[any] = None
    result: any = None

    def check(self, anno: RiskAnnotation):
        if self.func:
            if callable(self.func):
                result = self.func(anno, **self.args)
                if result == self.result:
                    return True
        return False


@dataclass
class RiskAnnotationList(object):
    items: List[RiskAnnotation] = field(default_factory=list)

    _i: int = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._i == len(self.items):
            self._i = 0
            raise StopIteration()
        anno = self.items[self._i]
        self._i += 1
        return anno

    def after(self, anno: RiskAnnotation):
        return get_annotations_after(self, anno)

    def filter(self, risk_type: RiskType = ""):
        current = self
        if risk_type:
            current = filter_annotations_by_type(current, risk_type)
        return current

    def find(self, risk_type: RiskType = "", condition: Union[FindCondition, List[FindCondition]] = None):
        return search_risk_annotations(self, risk_type, condition)


def get_annotations_after(anno_list: RiskAnnotationList, anno: RiskAnnotation):
    sub_list = []
    found = False
    for anno_i in anno_list:
        if anno_i.equal_to(anno):
            found = True
        if found:
            sub_list.append(anno_i)
    if not found:
        raise ValueError(f"Annotation {anno} is not found in the specified AnnotationList")
    return RiskAnnotationList(sub_list)


def filter_annotations_by_type(anno_list: RiskAnnotationList, risk_type: RiskType):
    sub_list = []
    for anno_i in anno_list:
        if anno_i.risk_type == risk_type:
            sub_list.append(anno_i)
    return sub_list


def search_risk_annotations(anno_list: RiskAnnotationList, risk_type: RiskType = "", condition: Union[FindCondition, List[FindCondition]] = None):
    matched = []
    for risk_anno in anno_list:
        if not isinstance(risk_anno, RiskAnnotation):
            continue
        if risk_type:
            if risk_anno.risk_type != risk_type:
                continue
        if condition:
            if isinstance(condition, FindCondition):
                condition = [condition]
            for cond in condition:
                if cond.check(risk_anno):
                    matched.append(risk_anno)
                    break
    return RiskAnnotationList(matched)


class ExecutableType:
    MODULE_TYPE = "Module"
    ROLE_TYPE = "Role"
    TASKFILE_TYPE = "TaskFile"


@dataclass
class BecomeInfo(object):
    enabled: bool = False
    become: str = ""
    user: str = ""
    method: str = ""
    flags: str = ""

    @staticmethod
    def from_options(options: dict):
        if "become" in options:
            become = options.get("become", "")
            enabled = False
            try:
                enabled = boolean(become)
            except Exception:
                pass
            user = options.get("become_user", "")
            method = options.get("become_method", "")
            flags = options.get("become_flags", "")
            return BecomeInfo(enabled=enabled, user=user, method=method, flags=flags)
        return None


@dataclass
class Task(Object, Resolvable):
    type: str = "task"
    name: str = ""
    module: str = ""
    index: int = -1
    play_index: int = -1
    defined_in: str = ""
    key: str = ""
    local_key: str = ""
    role: str = ""
    collection: str = ""
    become: BecomeInfo = None
    variables: dict = field(default_factory=dict)
    registered_variables: dict = field(default_factory=dict)
    set_facts: dict = field(default_factory=dict)
    loop: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)
    module_options: dict = field(default_factory=dict)
    executable: str = ""
    executable_type: str = ""
    collections_in_play: list = field(default_factory=list)

    yaml_lines: str = ""
    line_num_in_file: list = field(default_factory=list)  # [begin, end]

    # FQCN for Module and Role. Or a file path for TaskFile.  resolved later
    resolved_name: str = ""
    # candidates of resovled_name
    possible_candidates: list = field(default_factory=list)

    def set_yaml_lines(self, fullpath="", yaml_lines="", task_name="", module_name="", module_options=None):
        if not task_name and not module_options:
            return
        found_line_num = -1
        lines = []
        if yaml_lines:
            lines = yaml_lines.splitlines()
        else:
            lines = open(fullpath, "r").read().splitlines()
        for i, line in enumerate(lines):
            if task_name:
                if task_name in line:
                    found_line_num = i
                    break
            elif "{}:".format(module_name) in line:
                if isinstance(module_options, str):
                    if module_options in line:
                        found_line_num = i
                        break
                elif isinstance(module_options, dict):
                    option_matched = False
                    for key in module_options:
                        if "{}:".format(key) in lines[i + 1]:
                            option_matched = True
                            break
                    if option_matched:
                        found_line_num = i
                        break
        if found_line_num < 0:
            return
        found_line = lines[found_line_num]
        is_top_of_block = found_line.replace(" ", "").startswith("-")
        begin_line_num = found_line_num
        indent_of_block = -1
        if is_top_of_block:
            indent_of_block = len(found_line.split("-")[0])
        else:
            found = False
            found_line = ""
            _indent_of_block = -1
            parts = found_line.split(" ")
            for i, p in enumerate(parts):
                if p != "":
                    break
                _indent_of_block = i + 1
            for i in range(len(lines)):
                index = begin_line_num
                _line = lines[index]
                is_top_of_block = _line.replace(" ", "").startswith("-")
                if is_top_of_block:
                    _indent = len(_line.split("-")[0])
                    if _indent < _indent_of_block:
                        found = True
                        found_line = _line
                        break
                begin_line_num -= 1
                if begin_line_num < 0:
                    break
            if not found:
                return
            indent_of_block = len(found_line.split("-")[0])
        index = begin_line_num + 1
        end_found = False
        end_line_num = -1
        for i in range(len(lines)):
            if index >= len(lines):
                break
            _line = lines[index]
            is_top_of_block = _line.replace(" ", "").startswith("-")
            if is_top_of_block:
                _indent = len(_line.split("-")[0])
                if _indent <= indent_of_block:
                    end_found = True
                    end_line_num = index - 1
                    break
            index += 1
            if index >= len(lines):
                end_found = True
                end_line_num = index
                break
        if not end_found:
            return
        if begin_line_num < 0 or end_line_num > len(lines) or begin_line_num > end_line_num:
            return
        self.yaml_lines = "\n".join(lines[begin_line_num : end_line_num + 1])
        self.line_num_in_file = [begin_line_num + 1, end_line_num + 1]
        return

    # this keeps original contents like comments, indentation
    # and quotes for string as much as possible
    def yaml(self, original_module=""):
        task_data = None
        try:
            task_data_wrapper = ariyaml.load(self.yaml_lines)
            task_data = task_data_wrapper[0]
        except Exception:
            pass

        if not task_data:
            return self.yaml_lines

        # task name
        if self.name:
            task_data["name"] = self.name
        elif "name" in task_data:
            task_data.pop("name")

        # module name
        if original_module:
            mo = deepcopy(task_data[original_module])
            task_data[self.module] = mo
        elif self.module and self.module not in task_data:
            task_data[self.module] = self.module_options

        # module options
        if isinstance(self.module_options, dict):
            current_mo = task_data[self.module]
            # if the module options was an old style inline parameter in YAML,
            # we can ignore them here because it is parsed as self.module_options
            if not isinstance(current_mo, dict):
                current_mo = {}
            old_keys = list(current_mo.keys())
            new_keys = list(self.module_options.keys())
            for old_key in old_keys:
                if old_key not in new_keys:
                    current_mo.pop(old_key)
            recursive_copy_dict(self.module_options, current_mo)
            task_data[self.module] = current_mo

        # task options
        if isinstance(self.options, dict):
            current_to = task_data
            old_keys = list(current_to.keys())
            new_keys = list(self.options.keys())
            for old_key in old_keys:
                if old_key in ["name", self.module]:
                    continue
                if old_key not in new_keys:
                    current_to.pop(old_key)
            options_without_name = {k: v for k, v in self.options.items() if k != "name"}
            recursive_copy_dict(options_without_name, current_to)
        task_data_wrapper[0] = current_to
        new_yaml = ariyaml.dump(task_data_wrapper)
        return new_yaml

    # this makes a yaml from task contents such as spec.module,
    # spec.options, spec.module_options in a fixed format
    # NOTE: this will lose comments and indentations in the original YAML
    def formatted_yaml(self):
        task_data = {}
        if self.name:
            task_data["name"] = self.name
        if self.module:
            task_data[self.module] = self.module_options
        for key, val in self.options.items():
            if key == "name":
                continue
            task_data[key] = val
        task_data = self.str2double_quoted_scalar(task_data)
        data = [task_data]
        return ariyaml.dump(data)

    def str2double_quoted_scalar(self, v):
        if isinstance(v, dict):
            for key, val in v.items():
                new_val = self.str2double_quoted_scalar(val)
                v[key] = new_val
        elif isinstance(v, list):
            for i, val in enumerate(v):
                new_val = self.str2double_quoted_scalar(val)
                v[i] = new_val
        elif isinstance(v, str):
            v = DoubleQuotedScalarString(v)
        else:
            pass
        return v

    def set_key(self, parent_key="", parent_local_key=""):
        set_task_key(self, parent_key, parent_local_key)

    def children_to_key(self):
        return self

    @property
    def defined_vars(self):
        d_vars = self.variables
        d_vars.update(self.registered_variables)
        d_vars.update(self.set_facts)
        return d_vars

    @property
    def tags(self):
        return self.options.get("tags", None)

    @property
    def when(self):
        return self.options.get("when", None)

    @property
    def action(self):
        return self.executable

    @property
    def resolved_action(self):
        return self.resolved_name

    @property
    def line_number(self):
        return self.line_num_in_file

    @property
    def id(self):
        return json.dumps(
            {
                "path": self.defined_in,
                "index": self.index,
                "play_index": self.play_index,
            }
        )

    @property
    def resolver_targets(self):
        return None


@dataclass
class MutableContent(object):
    _yaml: str = ""
    _task_spec: Task = None

    @staticmethod
    def from_task_spec(task_spec):
        mc = MutableContent(
            _yaml=task_spec.yaml_lines,
            _task_spec=deepcopy(task_spec),
        )
        return mc

    def omit_task_name(self):
        # if `name` is None or empty string, Task.yaml() won't output the field
        self._task_spec.name = None
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        return self

    def set_module_name(self, module_name):
        original_module = deepcopy(self._task_spec.module)
        self._task_spec.module = module_name
        self._yaml = self._task_spec.yaml(original_module=original_module)
        self._task_spec.yaml_lines = self._yaml
        return self

    def replace_key(self, old_key: str, new_key: str):
        if old_key in self._task_spec.options:
            value = self._task_spec.options[old_key]
            self._task_spec.options.pop(old_key)
            self._task_spec.options[new_key] = value
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        return self

    def replace_value(self, old_value: str, new_value: str):
        original_new_value = deepcopy(new_value)
        need_restore = False
        keys_to_be_restored = []
        if isinstance(new_value, str):
            new_value = DoubleQuotedScalarString(new_value)
            need_restore = True
        for k, v in self._task_spec.options.items():
            if type(v) != type(old_value):
                continue
            if v != old_value:
                continue
            self._task_spec.options[k] = new_value
            keys_to_be_restored.append(k)
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        if need_restore:
            for k, v in self._task_spec.options.items():
                if k in keys_to_be_restored:
                    self._task_spec.options[k] = original_new_value
        return self

    def remove_key(self, key):
        if key in self._task_spec.options:
            self._task_spec.options.pop(key)
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        return self

    def set_new_module_arg_key(self, key, value):
        original_value = deepcopy(value)
        need_restore = False
        if isinstance(value, str):
            value = DoubleQuotedScalarString(value)
            need_restore = True
        self._task_spec.module_options[key] = value
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        if need_restore:
            self._task_spec.module_options[key] = original_value
        return self

    def remove_module_arg_key(self, key):
        if key in self._task_spec.module_options:
            self._task_spec.module_options.pop(key)
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        return self

    def replace_module_arg_key(self, old_key: str, new_key: str):
        if old_key in self._task_spec.module_options:
            value = self._task_spec.module_options[old_key]
            self._task_spec.module_options.pop(old_key)
            self._task_spec.module_options[new_key] = value
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        return self

    def replace_module_arg_value(self, key: str = "", old_value: any = None, new_value: any = None):
        original_new_value = deepcopy(new_value)
        need_restore = False
        keys_to_be_restored = []
        if isinstance(new_value, str):
            new_value = DoubleQuotedScalarString(new_value)
            need_restore = True
        for k in self._task_spec.module_options:
            # if `key` is specified, skip other keys
            if key and k != key:
                continue
            value = self._task_spec.module_options[k]
            if type(value) == type(old_value) and value == old_value:
                self._task_spec.module_options[k] = new_value
                keys_to_be_restored.append(k)
        self._yaml = self._task_spec.yaml()
        self._task_spec.yaml_lines = self._yaml
        if need_restore:
            for k in self._task_spec.module_options:
                if k in keys_to_be_restored:
                    self._task_spec.module_options[k] = original_new_value
        return self

    def replace_with_dict(self, new_dict: dict):
        # import this here to avoid circular import
        from .model_loader import load_task

        yaml_lines = ariyaml.dump([new_dict])
        new_task = load_task(
            path=self._task_spec.defined_in,
            index=self._task_spec.index,
            task_block_dict=new_dict,
            role_name=self._task_spec.role,
            collection_name=self._task_spec.collection,
            collections_in_play=self._task_spec.collections_in_play,
            play_index=self._task_spec.play_index,
            yaml_lines=yaml_lines,
        )
        self._yaml = yaml_lines
        self._task_spec = new_task
        return self

    def replace_module_arg_with_dict(self, new_dict: dict):
        self._task_spec.module_options = new_dict
        self._yaml = self._task_spec.yaml()
        return self

    # this keeps original contents like comments, indentation
    # and quotes for string as much as possible
    def yaml(self):
        return self._yaml

    # this makes a yaml from task contents such as spec.module,
    # spec.options, spec.module_options in a fixed format
    # NOTE: this will lose comments and indentations in the original YAML
    def formatted_yaml(self):
        return self._task_spec.formatted_yaml()


@dataclass
class TaskCall(CallObject, RunTarget):
    type: str = "taskcall"
    # annotations are used for storing generic analysis data
    # any Annotators in "annotators" dir can add them to this object
    annotations: List[Annotation] = field(default_factory=list)
    args: Arguments = field(default_factory=Arguments)
    variable_set: dict = field(default_factory=dict)
    variable_use: dict = field(default_factory=dict)
    become: BecomeInfo = None

    module: Module = None
    content: MutableContent = None

    def get_annotation_by_type(self, type_str=""):
        matched = [an for an in self.annotations if an.type == type_str]
        return matched

    def get_annotation_by_type_and_attr(self, type_str="", key="", val=None):
        matched = [an for an in self.annotations if hasattr(an, "type") and an.type == type_str and getattr(an, key, None) == val]
        return matched

    def set_annotation(self, key: str, value: any, rule_id: str):
        end_to_set = False
        for an in self.annotations:
            if not hasattr(an, "key"):
                continue
            if getattr(an, "key") == key:
                setattr(an, "value", value)
                end_to_set = True
                break
        if not end_to_set:
            self.annotations.append(Annotation(key=key, value=value, rule_id=rule_id))
        return

    def get_annotation(self, key: str, __default: any = None, rule_id: str = ""):
        value = __default
        for an in self.annotations:
            if not hasattr(an, "key"):
                continue
            if rule_id:
                if hasattr(an, "rule_id"):
                    if an.rule_id != rule_id:
                        continue
            if getattr(an, "key") == key:
                value = getattr(an, "value", __default)
                break
        return value

    def has_annotation_by_condition(self, cond: AnnotationCondition):
        anno = self.get_annotation_by_condition(cond)
        if anno:
            return True
        return False

    def get_annotation_by_condition(self, cond: AnnotationCondition):
        _annotations = self.annotations
        if cond.type:
            _annotations = [an for an in _annotations if an.type == RiskAnnotation.type and an.risk_type == cond.type]
        if cond.attr_conditions:
            for (key, val) in cond.attr_conditions:
                _annotations = [an for an in _annotations if hasattr(an, key) and getattr(an, key) == val]
        if _annotations:
            return _annotations[0]
        return None

    def file_info(self):
        file = self.spec.defined_in
        lines = "?"
        if len(self.spec.line_number) == 2:
            l_num = self.spec.line_number
            lines = f"L{l_num[0]}-{l_num[1]}"
        return file, lines

    @property
    def resolved_name(self):
        return self.spec.resolved_name

    @property
    def resolved_action(self):
        return self.resolved_name

    @property
    def action_type(self):
        return self.spec.executable_type


@dataclass
class AnsibleRunContext(object):
    sequence: RunTargetList = None
    root_key: str = ""

    # used by rule check
    current: RunTarget = None
    _i: int = 0

    # TODO: implement the following attributes
    vars: any = None
    host_info: any = None

    def __len__(self):
        return len(self.sequence)

    def __iter__(self):
        return self

    def __next__(self):
        if self._i == len(self.sequence):
            self._i = 0
            self.current = None
            raise StopIteration()
        t = self.sequence[self._i]
        self.current = t
        self._i += 1
        return t

    def __getitem__(self, i):
        return self.sequence[i]

    @staticmethod
    def from_tree(tree: ObjectList):
        if not tree:
            return AnsibleRunContext()
        if len(tree.items) == 0:
            return AnsibleRunContext()

        root_key = tree.items[0].spec.key
        sequence_items = []
        for item in tree.items:
            if not isinstance(item, RunTarget):
                continue
            sequence_items.append(item)
        tl = RunTargetList(items=sequence_items)
        return AnsibleRunContext(sequence=tl, root_key=root_key)

    @staticmethod
    def from_targets(targets: List[RunTarget], root_key: str = ""):
        if not root_key:
            if len(targets) > 0:
                root_key = targets[0].spec.key
        tl = RunTargetList(items=targets)
        return AnsibleRunContext(sequence=tl, root_key=root_key)

    def find(self, target: RunTarget):
        for t in self.sequence:
            if t.key == target.key:
                return t
        return None

    def before(self, target: RunTarget):
        targets = []
        for rt in self.sequence:
            if rt.key == target.key:
                break
            targets.append(rt)
        return AnsibleRunContext.from_targets(targets, root_key=self.root_key)

    def search(self, cond: AnnotationCondition):
        targets = [t for t in self.sequence if t.type == RunTargetType.Task and t.has_annotation_by_condition(cond)]
        return AnsibleRunContext.from_targets(targets, root_key=self.root_key)

    def is_end(self, target: RunTarget):
        if len(self) == 0:
            return False
        return target.key == self.sequence[-1].key

    def is_begin(self, target: RunTarget):
        if len(self) == 0:
            return False
        return target.key == self.sequence[0].key

    def copy(self):
        return AnsibleRunContext.from_targets(targets=self.sequence.items, root_key=self.root_key)

    @property
    def info(self):
        if not self.root_key:
            return {}
        return get_obj_info_by_key(self.root_key)

    @property
    def taskcalls(self):
        return [t for t in self.sequence if t.type == RunTargetType.Task]

    @property
    def tasks(self):
        return self.taskcalls

    @property
    def annotations(self):
        anno_list = []
        for tc in self.taskcalls:
            anno_list.extend(tc.annotations)
        return RiskAnnotationList(anno_list)


@dataclass
class TaskFile(Object, Resolvable):
    type: str = "taskfile"
    name: str = ""
    defined_in: str = ""
    key: str = ""
    local_key: str = ""
    tasks: list = field(default_factory=list)
    # role name of this task file
    # this might be empty because a task file can be defined out of roles
    role: str = ""
    collection: str = ""

    yaml_lines: str = ""

    used_in: list = field(default_factory=list)  # resolved later

    annotations: dict = field(default_factory=dict)

    variables: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)

    def set_key(self):
        set_taskfile_key(self)

    def children_to_key(self):
        task_keys = [t.key if isinstance(t, Task) else t for t in self.tasks]
        self.tasks = task_keys
        return self

    @property
    def resolver_targets(self):
        return self.tasks


@dataclass
class TaskFileCall(CallObject, RunTarget):
    type: str = "taskfilecall"


@dataclass
class Role(Object, Resolvable):
    type: str = "role"
    name: str = ""
    defined_in: str = ""
    key: str = ""
    local_key: str = ""
    fqcn: str = ""
    metadata: dict = field(default_factory=dict)
    collection: str = ""
    playbooks: list = field(default_factory=list)
    # 1 role can have multiple task yamls
    taskfiles: list = field(default_factory=list)
    # roles/xxxx/library/zzzz.py can be called as module zzzz
    modules: list = field(default_factory=list)
    dependency: dict = field(default_factory=dict)
    requirements: dict = field(default_factory=dict)

    source: str = ""  # collection/scm repo/galaxy

    annotations: dict = field(default_factory=dict)

    default_variables: dict = field(default_factory=dict)
    variables: dict = field(default_factory=dict)
    # key: loop_var (default "item"), value: list/dict of item value
    loop: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)

    def set_key(self):
        set_role_key(self)

    def children_to_key(self):
        module_keys = [m.key if isinstance(m, Module) else m for m in self.modules]
        self.modules = sorted(module_keys)

        playbook_keys = [p.key if isinstance(p, Playbook) else p for p in self.playbooks]
        self.playbooks = sorted(playbook_keys)

        taskfile_keys = [tf.key if isinstance(tf, TaskFile) else tf for tf in self.taskfiles]
        self.taskfiles = sorted(taskfile_keys)
        return self

    @property
    def resolver_targets(self):
        return self.taskfiles + self.modules


@dataclass
class RoleCall(CallObject, RunTarget):
    type: str = "rolecall"


@dataclass
class RoleInPlay(Object, Resolvable):
    type: str = "roleinplay"
    name: str = ""
    options: dict = field(default_factory=dict)
    defined_in: str = ""
    role_index: int = -1
    play_index: int = -1

    role: str = ""
    collection: str = ""

    resolved_name: str = ""  # resolved later
    # candidates of resovled_name
    possible_candidates: list = field(default_factory=list)

    annotations: dict = field(default_factory=dict)
    collections_in_play: list = field(default_factory=list)

    @property
    def resolver_targets(self):
        return None


@dataclass
class RoleInPlayCall(CallObject):
    type: str = "roleinplaycall"


@dataclass
class Play(Object, Resolvable):
    type: str = "play"
    name: str = ""
    defined_in: str = ""
    index: int = -1
    key: str = ""
    local_key: str = ""

    role: str = ""
    collection: str = ""
    import_module: str = ""
    import_playbook: str = ""
    pre_tasks: list = field(default_factory=list)
    tasks: list = field(default_factory=list)
    post_tasks: list = field(default_factory=list)
    # not actual Role, but RoleInPlay defined in this playbook
    roles: list = field(default_factory=list)
    options: dict = field(default_factory=dict)
    collections_in_play: list = field(default_factory=list)
    become: BecomeInfo = None
    variables: dict = field(default_factory=dict)

    def set_key(self, parent_key="", parent_local_key=""):
        set_play_key(self, parent_key, parent_local_key)

    def children_to_key(self):
        pre_task_keys = [t.key if isinstance(t, Task) else t for t in self.pre_tasks]
        self.pre_tasks = pre_task_keys

        task_keys = [t.key if isinstance(t, Task) else t for t in self.tasks]
        self.tasks = task_keys

        post_task_keys = [t.key if isinstance(t, Task) else t for t in self.post_tasks]
        self.post_tasks = post_task_keys
        return self

    @property
    def id(self):
        return json.dumps({"path": self.defined_in, "index": self.index})

    @property
    def resolver_targets(self):
        return self.pre_tasks + self.tasks + self.roles


@dataclass
class PlayCall(CallObject, RunTarget):
    type: str = "playcall"


@dataclass
class Playbook(Object, Resolvable):
    type: str = "playbook"
    name: str = ""
    defined_in: str = ""
    key: str = ""
    local_key: str = ""

    yaml_lines: str = ""

    role: str = ""
    collection: str = ""

    plays: list = field(default_factory=list)

    used_in: list = field(default_factory=list)  # resolved later

    annotations: dict = field(default_factory=dict)

    variables: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)

    def set_key(self):
        set_playbook_key(self)

    def children_to_key(self):
        play_keys = [play.key if isinstance(play, Play) else play for play in self.plays]
        self.plays = play_keys
        return self

    @property
    def resolver_targets(self):
        if "plays" in self.__dict__:
            return self.plays
        else:
            return self.roles + self.tasks


@dataclass
class PlaybookCall(CallObject, RunTarget):
    type: str = "playbookcall"


class InventoryType:
    GROUP_VARS_TYPE = "group_vars"
    HOST_VARS_TYPE = "host_vars"
    UNKNOWN_TYPE = ""


@dataclass
class Inventory(JSONSerializable):
    type: str = "inventory"
    name: str = ""
    defined_in: str = ""
    inventory_type: str = ""
    group_name: str = ""
    host_name: str = ""
    variables: dict = field(default_factory=dict)


@dataclass
class Repository(Object, Resolvable):
    type: str = "repository"
    name: str = ""
    path: str = ""
    key: str = ""
    local_key: str = ""

    # if set, this repository is a collection repository
    my_collection_name: str = ""

    playbooks: list = field(default_factory=list)
    roles: list = field(default_factory=list)

    # for playbook scan
    target_playbook_path: str = ""

    # for taskfile scan
    target_taskfile_path: str = ""

    requirements: dict = field(default_factory=dict)

    installed_collections_path: str = ""
    installed_collections: list = field(default_factory=list)

    installed_roles_path: str = ""
    installed_roles: list = field(default_factory=list)
    modules: list = field(default_factory=list)
    taskfiles: list = field(default_factory=list)

    inventories: list = field(default_factory=list)

    version: str = ""

    annotations: dict = field(default_factory=dict)

    def set_key(self):
        set_repository_key(self)

    def children_to_key(self):
        module_keys = [m.key if isinstance(m, Module) else m for m in self.modules]
        self.modules = sorted(module_keys)

        playbook_keys = [p.key if isinstance(p, Playbook) else p for p in self.playbooks]
        self.playbooks = sorted(playbook_keys)

        taskfile_keys = [tf.key if isinstance(tf, TaskFile) else tf for tf in self.taskfiles]
        self.taskfiles = sorted(taskfile_keys)

        role_keys = [r.key if isinstance(r, Role) else r for r in self.roles]
        self.roles = sorted(role_keys)
        return self

    @property
    def resolver_targets(self):
        return self.playbooks + self.roles + self.modules + self.installed_roles + self.installed_collections


@dataclass
class RepositoryCall(CallObject):
    type: str = "repositorycall"


def call_obj_from_spec(spec: Object, caller: CallObject, index: int = 0):
    if isinstance(spec, Repository):
        return RepositoryCall.from_spec(spec, caller, index)
    elif isinstance(spec, Playbook):
        return PlaybookCall.from_spec(spec, caller, index)
    elif isinstance(spec, Play):
        return PlayCall.from_spec(spec, caller, index)
    elif isinstance(spec, RoleInPlay):
        return RoleInPlayCall.from_spec(spec, caller, index)
    elif isinstance(spec, Role):
        return RoleCall.from_spec(spec, caller, index)
    elif isinstance(spec, TaskFile):
        return TaskFileCall.from_spec(spec, caller, index)
    elif isinstance(spec, Task):
        taskcall = TaskCall.from_spec(spec, caller, index)
        taskcall.content = MutableContent.from_task_spec(task_spec=spec)
        return taskcall
    elif isinstance(spec, Module):
        return ModuleCall.from_spec(spec, caller, index)
    return None


# inherit Repository just for convenience
# this is not a Repository but one or multiple Role / Collection
@dataclass
class GalaxyArtifact(Repository):
    type: str = ""  # Role or Collection

    # make it easier to search a module
    module_dict: dict = field(default_factory=dict)
    # make it easier to search a task
    task_dict: dict = field(default_factory=dict)
    # make it easier to search a taskfile
    taskfile_dict: dict = field(default_factory=dict)
    # make it easier to search a role
    role_dict: dict = field(default_factory=dict)
    # make it easier to search a playbook
    playbook_dict: dict = field(default_factory=dict)
    # make it easier to search a collection
    collection_dict: dict = field(default_factory=dict)


# following ansible-lint severity levels
class Severity:
    VERY_HIGH = "very_high"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    VERY_LOW = "very_low"
    NONE = "none"


_severity_level_mapping = {
    Severity.VERY_HIGH: 5,
    Severity.HIGH: 4,
    Severity.MEDIUM: 3,
    Severity.LOW: 2,
    Severity.VERY_LOW: 1,
    Severity.NONE: 0,
}


class RuleTag:
    NETWORK = "network"
    COMMAND = "command"
    DEPENDENCY = "dependency"
    SYSTEM = "system"
    PACKAGE = "package"
    CODING = "coding"
    VARIABLE = "variable"
    QUALITY = "quality"
    DEBUG = "debug"


@dataclass
class RuleMetadata(object):
    rule_id: str = ""
    description: str = ""
    name: str = ""

    version: str = ""
    severity: str = ""
    tags: tuple = ()


@dataclass
class SpecMutation(object):
    key: str = None
    changes: list = field(default_factory=list)
    object: Object = field(default_factory=Object)
    rule: RuleMetadata = field(default_factory=RuleMetadata)


@dataclass
class RuleResult(object):
    rule: RuleMetadata = None

    verdict: bool = False
    detail: dict = None
    file: tuple = None
    error: str = None

    matched: bool = False
    duration: float = None

    def __post_init__(self):
        if self.verdict:
            self.verdict = True
        else:
            self.verdict = False

    def set_value(self, key: str, value: any):
        self.detail[key] = value

    def get_detail(self):
        return self.detail


@dataclass
class Rule(RuleMetadata):
    # `enabled` represents if the rule is enabled or not
    enabled: bool = False

    # `precedence` represents the order of the rule evaluation.
    # A rule with a lower number will be evaluated earlier than others.
    precedence: int = 10

    # `spec_mutation` represents if the rule mutates spec objects
    # if there are any spec mutations, re-run the scan later with the mutated spec
    spec_mutation: bool = False

    def __post_init__(self, rule_id: str = "", description: str = ""):
        if rule_id:
            self.rule_id = rule_id
        if description:
            self.description = description

        if not self.rule_id:
            raise ValueError("A rule must have a unique rule_id")

        if not self.description:
            raise ValueError("A rule must have a description")

    def match(self, ctx: AnsibleRunContext) -> bool:
        raise ValueError("this is a base class method")

    def process(self, ctx: AnsibleRunContext):
        raise ValueError("this is a base class method")

    def print(self, result: RuleResult):
        output = f"ruleID={self.rule_id}, severity={self.severity}, description={self.description}, result={result.verdict}"

        if result.file:
            output += f", file={result.file}"
        if result.detail:
            output += f", detail={result.detail}"
        return output

    def to_json(self, result: RuleResult):
        return json.dumps(result.detail)

    def error(self, result: RuleResult):
        if result.error:
            return result.error
        return None

    def get_metadata(self):
        return RuleMetadata(
            rule_id=self.rule_id,
            description=self.description,
            name=self.name,
            version=self.version,
            severity=self.severity,
            tags=self.tags,
        )


@dataclass
class NodeResult(JSONSerializable):
    node: RunTarget = None
    rules: List[RuleResult] = field(default_factory=list)

    def results(self):
        return self.rules

    def find_result(self, rule_id: str):
        filtered = [r for r in self.rules if r.rule.rule_id == rule_id]
        if not filtered:
            return None
        return filtered[0]

    def search_results(
        self,
        rule_id: Union[str, list] = None,
        tag: Union[str, list] = None,
        matched: bool = None,
        verdict: bool = None,
    ):
        if not rule_id and not tag:
            return self.rules

        filtered = self.rules
        if rule_id:
            target_rule_ids = []
            if isinstance(rule_id, str):
                target_rule_ids = [rule_id]
            elif isinstance(rule_id, list):
                target_rule_ids = rule_id
            filtered = [r for r in filtered if r.rule.rule_id in target_rule_ids]

        if tag:
            target_tags = []
            if isinstance(tag, str):
                target_tags = [tag]
            elif isinstance(tag, list):
                target_tags = tag
            filtered = [r for r in filtered for t in r.rule.tags if t in target_tags]

        if matched is not None:
            filtered = [r for r in filtered if r.matched == matched]

        if verdict is not None:
            filtered = [r for r in filtered if r.verdict == verdict]

        return filtered


@dataclass
class TargetResult(JSONSerializable):
    target_type: str = ""  # playbook, role or taskfile
    target_name: str = ""
    nodes: List[NodeResult] = field(default_factory=list)

    def applied_rules(self):
        results = []
        for n in self.nodes:
            matched_rules = n.search_results(matched=True)
            if matched_rules:
                results.extend()
        return results

    def matched_rules(self):
        results = []
        for n in self.nodes:
            matched_rules = n.search_results(verdict=True)
            if matched_rules:
                results.extend()
        return results

    def tasks(self):
        return self._filter(TaskCall)

    def task(self, name):
        return self._find_by_name(name)

    def roles(self):
        return self._filter(RoleCall)

    def role(self, name):
        return self._find_by_name(name)

    def playbooks(self):
        return self._filter(PlaybookCall)

    def playbook(self, name):
        return self._find_by_name(name)

    def plays(self):
        return self._filter(PlayCall)

    def play(self, name):
        return self._find_by_name(name)

    def taskfiles(self):
        return self._filter(TaskFileCall)

    def taskfile(self, name):
        return self._find_by_name(name)

    def _find_by_name(self, name):
        filtered_nodes = [nr for nr in self.nodes if nr.node.spec.name == name]
        if not filtered_nodes:
            return None
        return filtered_nodes[0]

    def _filter(self, type):
        filtered_nodes = [nr for nr in self.nodes if isinstance(nr.node, type)]
        return TargetResult(target_type=self.target_type, target_name=self.target_name, nodes=filtered_nodes)


@dataclass
class ARIResult(JSONSerializable):
    targets: List[TargetResult] = field(default_factory=list)

    def playbooks(self):
        return self._filter("playbook")

    def playbook(self, name="", path="", yaml_str=""):
        if name:
            return self._find_by_name(name)

        # TODO: use path correctly
        if path:
            name = os.path.basename(path)
            return self._find_by_name(name)

        if yaml_str:
            return self._find_by_yaml_str(yaml_str, "playbook")

        return None

    def roles(self):
        return self._filter("role")

    def role(self, name):
        return self._find_by_name(name)

    def taskfiles(self):
        return self._filter("taskfile")

    def taskfile(self, name="", path="", yaml_str=""):
        if name:
            return self._find_by_name(name)

        # TODO: use path correctly
        if path:
            name = os.path.basename(path)
            return self._find_by_name(name)

        if yaml_str:
            return self._find_by_yaml_str(yaml_str, "taskfile")

        return None

    def find_target(self, name="", path="", yaml_str="", target_type=""):
        if name:
            return self._find_by_name(name)

        # TODO: use path correctly
        if path:
            name = os.path.basename(path)
            return self._find_by_name(name)

        if yaml_str:
            return self._find_by_yaml_str(yaml_str, target_type)

        return None

    def _find_by_name(self, name):
        filtered_targets = [tr for tr in self.targets if tr.target_name == name]
        if not filtered_targets:
            return None
        return filtered_targets[0]

    def _find_by_yaml_str(self, yaml_str, type_str):
        type_only_result = self._filter(type_str)
        if not type_only_result:
            return None
        filtered_targets = [
            tr
            for tr in type_only_result.targets
            if tr.nodes and hasattr(tr.nodes[0].node.spec, "yaml_lines") and tr.nodes[0].node.spec.yaml_lines == yaml_str
        ]
        if not filtered_targets:
            return None
        return filtered_targets[0]

    def _filter(self, type_str):
        filtered_targets = [tr for tr in self.targets if tr.target_type == type_str]
        return ARIResult(targets=filtered_targets)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import re


# glob.glob() may cause infinite loop when there is symlink loop
# safe_glob() support the case by `followlink=False` option as default
def safe_glob(patterns, root_dir="", recursive=True, followlinks=False):
    pattern_list = []
    if isinstance(patterns, list):
        pattern_list = [p for p in patterns]
    elif isinstance(patterns, str):
        pattern_list = [patterns]
    else:
        raise ValueError("patterns for safe_glob() must be str or list of str")

    matched_files = []
    for pattern in pattern_list:
        # if root dir is not specified, automatically decide it with pattern
        # e.g.) pattern "testdir1/testdir2/*.py"
        #       --> root_dir "testdir1/testdir2"
        root_dir_for_this_pattern = ""
        if root_dir == "":
            root_cand = pattern.split("*")[0]
            if root_cand.endswith("/"):
                root_cand = root_cand[:-1]  # trim "/" suffix
            else:
                root_cand = "/".join(root_cand.split("/")[:-1])  # testdir1/testdir2/file-*.txt --> testdir1/testdir2
            root_dir_for_this_pattern = root_cand
        else:
            root_dir_for_this_pattern = root_dir

        # if recusive, use os.walk to search files recursively
        if recursive:
            for dirpath, folders, files in os.walk(root_dir_for_this_pattern, followlinks=followlinks):
                for file in files:
                    fpath = os.path.join(dirpath, file)
                    fpath = os.path.normpath(fpath)
                    if fpath in matched_files:
                        continue
                    if pattern_match(pattern, fpath):
                        matched_files.append(fpath)
        else:
            # otherwise, just use os.listdir to avoid
            # unnecessary loading time of os.walk
            files = os.listdir(root_dir_for_this_pattern)
            for file in files:
                fpath = os.path.join(root_dir, file)
                fpath = os.path.normpath(fpath)
                if fpath in matched_files:
                    continue
                if pattern_match(pattern, fpath):
                    matched_files.append(fpath)
    return matched_files


def pattern_match(pattern, fpath):
    pattern = pattern.replace("**/", "<ANY>")
    pattern = pattern.replace("*", "[^/]*")
    pattern = pattern.replace("<ANY>", ".*")
    regex_pattern = r"^{}$".format(pattern)
    return re.match(regex_pattern, fpath)


if __name__ == "__main__":
    dir, pattern = sys.argv[1], sys.argv[2]
    found = safe_glob(dir, [pattern], recursive=False)
    print(found)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import yaml
import subprocess
import tempfile
import glob
import re
import sys
import datetime
import tarfile
from dataclasses import dataclass, field, asdict

import ansible_risk_insight.logger as logger
from .models import (
    LoadType,
)
from .dependency_finder import find_dependency
from .utils import (
    escape_url,
    install_galaxy_target,
    install_github_target,
    get_installed_metadata,
    get_hash_of_url,
    is_url,
    is_local_path,
)
from .loader import (
    get_target_name,
    remove_subdirectories,
    trim_suffix,
)
from .safe_glob import safe_glob

collection_manifest_json = "MANIFEST.json"
collection_files_json = "FILES.json"
role_meta_main_yml = "meta/main.yml"
role_meta_main_yaml = "meta/main.yaml"
requirements_yml = "requirements.yml"

supported_target_types = [
    LoadType.PROJECT,
    LoadType.COLLECTION,
    LoadType.ROLE,
    LoadType.PLAYBOOK,
]

download_metadata_file = "download_meta.json"


@dataclass
class DownloadMetadata(object):
    name: str = ""
    type: str = ""
    version: str = ""
    author: str = ""
    download_url: str = ""
    download_src_path: str = ""  # path to put tar.gz
    hash: str = ""
    metafile_path: str = ""  # path to manifest.json/meta.yml
    files_json_path: str = ""
    download_timestamp: str = ""
    cache_enabled: bool = False
    cache_dir: str = ""  # path to put cache data
    source_repository: str = ""
    requirements_file: str = ""


@dataclass
class Dependency(object):
    dir: str = ""
    name: str = ""
    metadata: DownloadMetadata = field(default_factory=DownloadMetadata)


@dataclass
class DependencyDirPreparator(object):
    root_dir: str = ""
    source_repository: str = ""
    target_type: str = ""
    target_name: str = ""
    target_version: str = ""
    target_path: str = ""
    target_dependency_dir: str = ""
    target_path_mappings: dict = field(default_factory=dict)
    metadata: DownloadMetadata = field(default_factory=DownloadMetadata)
    download_location: str = ""
    dependency_dir_path: str = ""
    silent: bool = False
    do_save: bool = False
    tmp_install_dir: tempfile.TemporaryDirectory = None

    # -- out --
    dependency_dirs: list = field(default_factory=list)

    def prepare_dir(self, root_install=True, is_src_installed=False, cache_enabled=False, cache_dir=""):
        logger.debug("setup base dirs")
        self.setup_dirs(cache_enabled, cache_dir)
        logger.debug("prepare target dir")
        self.prepare_root_dir(root_install, is_src_installed)
        logger.debug("search dependencies")
        dependencies = find_dependency(self.target_type, self.target_path, self.target_dependency_dir)
        logger.debug("prepare dir for dependencies")
        self.prepare_dependency_dir(dependencies, cache_enabled, cache_dir)
        return self.dependency_dirs

    def setup_dirs(self, cache_enabled=False, cache_dir=""):
        self.download_location = os.path.join(self.root_dir, "archives")
        self.dependency_dir_path = self.root_dir
        # check download_location
        if not os.path.exists(self.download_location):
            os.makedirs(self.download_location)
        # check cache_dir
        if cache_enabled and not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        # check dependency_dir_path
        if not os.path.exists(self.dependency_dir_path):
            os.makedirs(self.dependency_dir_path)
        return

    def prepare_root_dir(self, root_install=True, is_src_installed=False):
        # install root
        if is_src_installed:
            pass
        else:
            # if a project target is a local path, then skip install
            if self.target_type in [LoadType.PROJECT, LoadType.PLAYBOOK, LoadType.TASKFILE] and not is_url(self.target_name):
                root_install = False

            # if a collection/role is a local path, then skip install (require MANIFEST.json or meta/main.yml to get the actual name)
            if self.target_type in [LoadType.COLLECTION, LoadType.ROLE] and is_local_path(self.target_name):
                root_install = False

            if root_install:
                self.src_install()
                if not self.silent:
                    logger.debug("install() done")
            else:
                download_url = ""
                version = ""
                hash = ""
                download_url, version = get_installed_metadata(self.target_type, self.target_name, self.target_path, self.target_dependency_dir)
                if download_url != "":
                    hash = get_hash_of_url(download_url)
                self.metadata.download_url = download_url
                self.metadata.version = version
                self.metadata.hash = hash
        return

    def prepare_dependency_dir(self, dependencies, cache_enabled=False, cache_dir=""):
        col_dependencies = dependencies.get("dependencies", {}).get("collections", [])
        role_dependencies = dependencies.get("dependencies", {}).get("roles", [])

        col_dependency_dirs = dependencies.get("paths", {}).get("collections", {})
        role_dependency_dirs = dependencies.get("paths", {}).get("roles", {})

        col_dependency_metadata = dependencies.get("metadata", {}).get("collections", {})
        # role_dependency_metadata = dependencies.get("metadata", {}).get("roles", {})

        # TODO: if requirements.yml is provided, download dependencies using it.

        for cdep in col_dependencies:
            col_name = cdep
            col_version = ""
            if type(cdep) is dict:
                col_name = cdep.get("name", "")
                col_version = cdep.get("version", "")
                if col_name == "":
                    col_name = cdep.get("source", "")

            logger.debug("prepare dir for {}:{}".format(col_name, col_version))
            downloaded_dep = Dependency(
                name=col_name,
            )
            downloaded_dep.metadata.type = LoadType.COLLECTION
            downloaded_dep.metadata.name = col_name
            downloaded_dep.metadata.cache_enabled = cache_enabled
            sub_dependency_dir_path = os.path.join(
                self.dependency_dir_path,
                "collections",
                "src",
            )

            if not os.path.exists(sub_dependency_dir_path):
                os.makedirs(sub_dependency_dir_path)

            if cache_enabled:
                logger.debug("cache enabled")
                # TODO: handle version
                is_exist, targz_file = self.is_download_file_exist(LoadType.COLLECTION, col_name, cache_dir)
                # check cache data
                if is_exist:
                    logger.debug("found cache data {}".format(targz_file))
                    metadata_file = os.path.join(targz_file.rsplit("/", 1)[0], download_metadata_file)
                    md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, col_name)
                    downloaded_dep.metadata = md
                else:
                    # if no cache data, download
                    logger.debug("cache data not found")
                    cache_location = os.path.join(cache_dir, "collection", col_name)
                    install_msg = self.download_galaxy_collection(col_name, cache_location, col_version, self.source_repository)
                    metadata = self.extract_collections_metadata(install_msg, cache_location)
                    metadata_file = self.export_data(metadata, cache_location, download_metadata_file)
                    md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, col_name)
                    downloaded_dep.metadata = md
                    if md:
                        targz_file = md.download_src_path
                # install collection from tar.gz
                self.install_galaxy_collection_from_targz(targz_file, sub_dependency_dir_path)
                downloaded_dep.metadata.cache_dir = targz_file
                parts = col_name.split(".")
                downloaded_dep.dir = os.path.join(sub_dependency_dir_path, "ansible_collections", parts[0], parts[1])
            elif col_name in col_dependency_dirs:
                logger.debug("use the specified dependency dirs")
                sub_dependency_dir_path = col_dependency_dirs[col_name]
                col_galaxy_data = col_dependency_metadata.get(col_name, {})
                if isinstance(col_galaxy_data, dict):
                    download_url = col_galaxy_data.get("download_url", "")
                    hash = ""
                    if download_url:
                        hash = get_hash_of_url(download_url)
                    version = col_galaxy_data.get("version", "")
                    downloaded_dep.metadata.source_repository = self.source_repository
                    downloaded_dep.metadata.download_url = download_url
                    downloaded_dep.metadata.hash = hash
                    downloaded_dep.metadata.version = version
                    downloaded_dep.dir = sub_dependency_dir_path
            else:
                logger.debug("download dependency {}".format(col_name))
                is_exist, targz = self.is_download_file_exist(
                    LoadType.COLLECTION, col_name, os.path.join(self.download_location, "collection", col_name)
                )
                if is_exist:
                    metadata_file = os.path.join(self.download_location, "collection", col_name, download_metadata_file)
                    self.install_galaxy_collection_from_targz(targz, sub_dependency_dir_path)
                    md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, col_name)
                else:
                    # check download_location
                    sub_download_location = os.path.join(self.download_location, "collection", col_name)
                    if not os.path.exists(sub_download_location):
                        os.makedirs(sub_download_location)
                    install_msg = self.download_galaxy_collection(col_name, sub_download_location, col_version, self.source_repository)
                    metadata = self.extract_collections_metadata(install_msg, sub_download_location)
                    metadata_file = self.export_data(metadata, sub_download_location, download_metadata_file)
                    md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, col_name)
                    if md:
                        self.install_galaxy_collection_from_reqfile(md.requirements_file, sub_dependency_dir_path)
                    # self.install_galaxy_collection_from_targz(md.download_src_path, sub_dependency_dir_path)
                if md is not None:
                    downloaded_dep.metadata = md
                downloaded_dep.metadata.source_repository = self.source_repository
                parts = col_name.split(".")
                downloaded_dep.dir = os.path.join(sub_dependency_dir_path, "ansible_collections", parts[0], parts[1])
            self.dependency_dirs.append(asdict(downloaded_dep))

        for rdep in role_dependencies:
            target_version = None
            if isinstance(rdep, dict):
                rdep_name = rdep.get("name", None)
                target_version = rdep.get("version", None)
                if not rdep_name:
                    rdep_name = rdep.get("role", None)
                rdep = rdep_name
            name = rdep
            if type(rdep) is dict:
                name = rdep.get("name", "")
                if name == "":
                    name = rdep.get("src", "")
            logger.debug("prepare dir for {}".format(name))
            downloaded_dep = Dependency(
                name=name,
            )
            downloaded_dep.metadata.type = LoadType.ROLE
            downloaded_dep.metadata.name = name
            downloaded_dep.metadata.cache_enabled = cache_enabled
            # sub_dependency_dir_path = "{}/{}".format(dependency_dir_path, rdep)

            sub_dependency_dir_path = os.path.join(
                self.dependency_dir_path,
                "roles",
                "src",
                name,
            )

            if not os.path.exists(sub_dependency_dir_path):
                os.makedirs(sub_dependency_dir_path)
            if cache_enabled:
                logger.debug("cache enabled")
                cache_dir_path = os.path.join(
                    cache_dir,
                    "roles",
                    "src",
                    name,
                )
                if os.path.exists(cache_dir_path) and len(os.listdir(cache_dir_path)) != 0:
                    logger.debug("cache data found")
                    metadata_file = os.path.join(cache_dir_path, download_metadata_file)
                    md = self.find_target_metadata(LoadType.ROLE, metadata_file, self.target_name)
                else:
                    logger.debug("cache data not found")
                    install_msg = install_galaxy_target(name, LoadType.ROLE, cache_dir_path, self.source_repository, target_version)
                    logger.debug("role install msg: {}".format(install_msg))
                    metadata = self.extract_roles_metadata(install_msg)
                    metadata_file = self.export_data(metadata, cache_dir_path, download_metadata_file)
                    md = self.find_target_metadata(LoadType.ROLE, metadata_file, self.target_name)
                self.move_src(sub_dependency_dir_path, cache_dir_path)
                if md is not None:
                    downloaded_dep.metadata = md
            elif name in role_dependency_dirs:
                logger.debug("use the specified dependency dirs")
                sub_dependency_dir_path = role_dependency_dirs[name]
            else:
                is_exist, _ = self.is_download_file_exist(LoadType.ROLE, name, os.path.join(self.download_location, "role", name))
                if is_exist:
                    metadata_file = os.path.join(self.download_location, "role", name, download_metadata_file)
                    md = self.find_target_metadata(LoadType.ROLE, metadata_file, name)
                    self.move_src(md.download_src_path, sub_dependency_dir_path)
                else:
                    install_msg = install_galaxy_target(name, LoadType.ROLE, sub_dependency_dir_path, self.source_repository)
                    logger.debug("role install msg: {}".format(install_msg))
                    metadata = self.extract_roles_metadata(install_msg)
                    sub_download_location = os.path.join(self.download_location, "role", name)
                    metadata_file = self.export_data(metadata, sub_download_location, download_metadata_file)
                    md = self.find_target_metadata(LoadType.ROLE, metadata_file, name)
                if md is not None:
                    downloaded_dep.metadata = md
            downloaded_dep.metadata.source_repository = self.source_repository
            downloaded_dep.dir = sub_dependency_dir_path
            self.dependency_dirs.append(asdict(downloaded_dep))
        return

    def src_install(self):
        try:
            self.setup_tmp_dir()
            self.root_install(self.tmp_install_dir)
        finally:
            self.clean_tmp_dir()
        return

    def root_install(self, tmp_src_dir):
        tmp_src_dir = os.path.join(self.tmp_install_dir.name, "src")
        if not os.path.exists(tmp_src_dir):
            os.makedirs(tmp_src_dir)

        logger.debug("root type is {}".format(self.target_type))
        if self.target_type == LoadType.PROJECT:
            # install_type = "github"
            # ansible-galaxy install
            if not self.silent:
                print("cloning {} from github".format(self.target_name))
            install_msg = install_github_target(self.target_name, tmp_src_dir)
            if not self.silent:
                logger.debug("STDOUT: {}".format(install_msg))
            # if self.target_dependency_dir == "":
            #     raise ValueError("dependency dir is required for project type")
            dependency_dir = self.target_dependency_dir
            dst_src_dir = os.path.join(self.target_path_mappings["src"], escape_url(self.target_name))
            self.metadata.download_url = self.target_name
        elif self.target_type == LoadType.COLLECTION:
            install_msg = ""
            sub_download_location = os.path.join(self.download_location, "collection", self.target_name)
            is_exist, targz_file = self.is_download_file_exist(LoadType.COLLECTION, self.target_name, self.download_location)
            if is_exist:
                metadata_file = os.path.join(targz_file.rsplit("/", 1)[0], download_metadata_file)
                md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, self.target_name)
            else:
                install_msg = self.download_galaxy_collection(self.target_name, sub_download_location, version=self.target_version)
                metadata = self.extract_collections_metadata(install_msg, sub_download_location)
                metadata_file = self.export_data(metadata, sub_download_location, download_metadata_file)
                md = self.find_target_metadata(LoadType.COLLECTION, metadata_file, self.target_name)
            if md:
                self.install_galaxy_collection_from_reqfile(md.requirements_file, tmp_src_dir)
            dst_src_dir = self.target_path_mappings["src"]
            dependency_dir = tmp_src_dir
            self.metadata = md
        elif self.target_type == LoadType.ROLE:
            install_msg = ""
            sub_download_location = os.path.join(self.download_location, "role", self.target_name)
            if os.path.exists(sub_download_location) and len(os.listdir(sub_download_location)) != 0:
                logger.debug("found cache data {}".format(sub_download_location))
                metadata_file = os.path.join(sub_download_location, download_metadata_file)
                md = self.find_target_metadata(LoadType.ROLE, metadata_file, self.target_name)
                self.move_src(sub_download_location, tmp_src_dir)
            else:
                install_msg = install_galaxy_target(
                    self.target_name, self.target_type, tmp_src_dir, self.source_repository, target_version=self.target_version
                )
                logger.debug("role install msg: {}".format(install_msg))
                metadata = self.extract_roles_metadata(install_msg)
                metadata_file = self.export_data(metadata, sub_download_location, download_metadata_file)
                md = self.find_target_metadata(LoadType.ROLE, metadata_file, self.target_name)
            if not md:
                raise ValueError("failed to install {} {}".format(self.target_type, self.target_name))
            dependency_dir = tmp_src_dir
            dst_src_dir = self.target_path_mappings["src"]
            self.metadata.download_src_path = "{}.{}".format(dst_src_dir, self.target_name)
            self.metadata = md
        else:
            raise ValueError("unsupported container type")

        self.install_log = install_msg
        if self.do_save:
            self.__save_install_log()

        self.set_index(dependency_dir)

        if not self.silent:
            print("moving index")
            logger.debug("index: {}".format(json.dumps(self.index)))
        if self.do_save:
            self.__save_index()
        if not os.path.exists(dst_src_dir):
            os.makedirs(dst_src_dir)
        self.move_src(tmp_src_dir, dst_src_dir)
        root_dst_src_path = "{}/{}".format(dst_src_dir, self.target_name)
        if self.target_type == LoadType.ROLE:
            self.update_role_download_src(metadata_file, dst_src_dir)
            self.metadata.download_src_path = root_dst_src_path
            self.metadata.metafile_path, _ = self.get_metafile_in_target(self.target_type, root_dst_src_path)
            self.metadata.author = self.get_author(self.target_type, self.metadata.metafile_path)

        if self.target_type == LoadType.PROJECT and self.target_dependency_dir:
            dst_dependency_dir = self.target_path_mappings["dependencies"]
            if not os.path.exists(dst_dependency_dir):
                os.makedirs(dst_dependency_dir)
            self.move_src(dependency_dir, dst_dependency_dir)
            logger.debug("root metadata: {}".format(json.dumps(asdict(self.metadata))))
        return

    def set_index(self, path):
        if not self.silent:
            print("crawl content")
        dep_type = LoadType.UNKNOWN
        target_path_list = []
        if os.path.isfile(path):
            # need further check?
            dep_type = LoadType.PLAYBOOK
            target_path_list.append = [path]
        elif os.path.exists(os.path.join(path, collection_manifest_json)):
            dep_type = LoadType.COLLECTION
            target_path_list = [path]
        elif os.path.exists(os.path.join(path, role_meta_main_yml)):
            dep_type = LoadType.ROLE
            target_path_list = [path]
        else:
            dep_type, target_path_list = find_ext_dependencies(path)

        if not self.silent:
            logger.info('the detected target type: "{}", found targets: {}'.format(self.target_type, len(target_path_list)))

        if self.target_type not in supported_target_types:
            logger.error("this target type is not supported")
            sys.exit(1)

        list = []
        for target_path in target_path_list:
            ext_name = get_target_name(dep_type, target_path)
            list.append(
                {
                    "name": ext_name,
                    "type": dep_type,
                }
            )

        index_data = {
            "dependencies": list,
            "path_mappings": self.target_path_mappings,
        }

        self.index = index_data

    def download_galaxy_collection(self, target, output_dir, version="", source_repository=""):
        server_option = ""
        if source_repository:
            server_option = "--server {}".format(source_repository)
        target_version = target
        if version:
            target_version = "{}:{}".format(target, version)
        logger.debug("downloading: {}".format("ansible-galaxy collection download '{}' {} -p {}".format(target_version, server_option, output_dir)))
        proc = subprocess.run(
            "ansible-galaxy collection download '{}' {} -p {}".format(target_version, server_option, output_dir),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        install_msg = proc.stdout
        logger.debug("STDOUT: {}".format(install_msg))
        return install_msg

    def download_galaxy_collection_from_reqfile(self, requirements, output_dir, source_repository=""):
        server_option = ""
        if source_repository:
            server_option = "--server {}".format(source_repository)
        proc = subprocess.run(
            "ansible-galaxy collection download -r {} {} -p {}".format(requirements, server_option, output_dir),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        install_msg = proc.stdout
        logger.debug("STDOUT: {}".format(install_msg))
        # return proc.stdout

    def install_galaxy_collection_from_targz(self, tarfile, output_dir):
        logger.debug("install collection from {}".format(tarfile))
        proc = subprocess.run(
            "ansible-galaxy collection install {} -p {}".format(tarfile, output_dir),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        install_msg = proc.stdout
        logger.debug("STDOUT: {}".format(install_msg))
        # return proc.stdout

    def install_galaxy_collection_from_reqfile(self, requirements, output_dir):
        logger.debug("install collection from {}".format(requirements))
        src_dir = requirements.replace(requirements_yml, "")
        proc = subprocess.run(
            "cd {} && ansible-galaxy collection install -r {} -p {}".format(src_dir, requirements, output_dir),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        install_msg = proc.stdout
        logger.debug("STDOUT: {}".format(install_msg))
        # return proc.stdout

    def is_download_file_exist(self, type, target, dir):
        is_exist = False
        filename = ""
        download_metadata_files = glob.glob(os.path.join(dir, type, target, "**", download_metadata_file), recursive=True)
        # check if tar.gz file already exists
        if len(download_metadata_files) != 0:
            for metafile in download_metadata_files:
                md = self.find_target_metadata(type, metafile, target)
                if md is not None:
                    is_exist = True
                    filename = md.download_src_path
        else:
            if os.path.exists(dir):
                namepart = target.replace(".", "-")
                for file in os.listdir(dir):
                    if file.endswith(".tar.gz") and namepart in file:
                        is_exist = True
                        filename = file
        return is_exist, filename

    def install_galaxy_role_from_reqfile(self, file, output_dir):
        proc = subprocess.run(
            "ansible-galaxy role install -r {} -p {}".format(file, output_dir),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        install_msg = proc.stdout
        logger.debug("STDOUT: {}".format(install_msg))

    def extract_collections_metadata(self, log_message, download_location):
        # -- log message
        # Downloading collection 'community.rabbitmq:1.2.3' to
        # Downloading https://galaxy.ansible.com/download/ansible-posix-1.4.0.tar.gz to ...
        download_url_pattern = r"Downloading (.*) to"
        url = ""
        version = ""
        hash = ""
        match_messages = re.findall(download_url_pattern, log_message)
        metadata_list = []
        for m in match_messages:
            metadata = DownloadMetadata()
            metadata.type = LoadType.COLLECTION
            if m.endswith("tar.gz"):
                logger.debug("extracted url from download log message: {}".format(m))
                url = m
                version = url.split("-")[-1].replace(".tar.gz", "")
                name = "{}.{}".format(url.split("-")[0].split("/")[-1], url.split("-")[1])
                metadata.download_url = url
                metadata.version = version
                metadata.name = name
                filename = url.split("/")[-1]
                fullpath = "{}/{}".format(download_location, filename)
                if not os.path.exists(fullpath):
                    logger.warning("failed to get metadata for {}".format(url))
                    pass
                m_time = os.path.getmtime(fullpath)
                dt_m = datetime.datetime.utcfromtimestamp(m_time).isoformat()
                metadata.download_timestamp = dt_m
                metadata.download_src_path = fullpath
                metadata.metafile_path, metadata.files_json_path = self.get_metafile_in_target(LoadType.COLLECTION, fullpath)
                metadata.author = self.get_author(LoadType.COLLECTION, metadata.metafile_path)
                metadata.requirements_file = "{}/{}".format(download_location, requirements_yml)

                if url != "":
                    hash = get_hash_of_url(url)
                    metadata.hash = hash
                logger.debug("metadata: {}".format(json.dumps(asdict(metadata))))

                metadata_list.append(asdict(metadata))
        result = {"collections": metadata_list}
        return result

    def extract_roles_metadata(self, log_message):
        # - downloading role from https://github.com/rhythmictech/ansible-role-awscli/archive/1.0.3.tar.gz
        # - extracting rhythmictech.awscli to /private/tmp/role-test/rhythmictech.awscli
        url = ""
        version = ""
        hash = ""
        metadata_list = []
        messages = log_message.splitlines()
        for i, line in enumerate(messages):
            if line.startswith("- downloading role from "):
                metadata = DownloadMetadata()
                metadata.type = LoadType.ROLE
                url = line.split(" ")[-1]
                logger.debug("extracted url from download log message: {}".format(url))
                version = url.split("/")[-1].replace(".tar.gz", "")
                name = messages[i + 1].split("/")[-1]
                metadata.download_url = url
                metadata.version = version
                metadata.name = name
                role_dir = messages[i + 1].split(" ")[-1]
                m_time = os.path.getmtime(role_dir)
                dt_m = datetime.datetime.utcfromtimestamp(m_time).isoformat()
                metadata.download_timestamp = dt_m
                metadata.download_src_path = role_dir
                if url != "":
                    hash = get_hash_of_url(url)
                    metadata.hash = hash
                logger.debug("metadata: {}".format(json.dumps(asdict(metadata))))
                metadata_list.append(asdict(metadata))
        result = {"roles": metadata_list}
        return result

    def find_target_metadata(self, type, metadata_file, target):
        with open(metadata_file, "r") as f:
            metadata = json.load(f)
        if type == LoadType.COLLECTION:
            metadata_list = metadata.get("collections", [])
        elif type == LoadType.ROLE:
            metadata_list = metadata.get("roles", [])
        else:
            logger.warning("metadata not found: unsupported type {} {}".format(type, target))
            return None
        for data in metadata_list:
            dm = DownloadMetadata(**data)
            if dm.name == target:
                logger.debug("found metadata: {}".format(target))
                return dm
        logger.warning("metadata not found: {}".format(target))
        return None

    def existing_dependency_dir_loader(self, dependency_type, dependency_dir_path):
        search_dirs = []
        if dependency_type == LoadType.COLLECTION:
            base_dir = dependency_dir_path
            if os.path.exists(os.path.join(dependency_dir_path, "ansible_collections")):
                base_dir = os.path.join(dependency_dir_path, "ansible_collections")
            namespaces = [ns for ns in os.listdir(base_dir) if not ns.endswith(".info")]
            for ns in namespaces:
                colls = [{"name": f"{ns}.{name}", "path": os.path.join(base_dir, ns, name)} for name in os.listdir(os.path.join(base_dir, ns))]
                search_dirs.extend(colls)

        dependency_dirs = []
        for dep_info in search_dirs:
            downloaded_dep = {"dir": "", "metadata": {}}
            downloaded_dep["dir"] = dep_info["path"]
            # meta data
            downloaded_dep["metadata"]["type"] = LoadType.COLLECTION
            downloaded_dep["metadata"]["name"] = dep_info["name"]
            dependency_dirs.append(downloaded_dep)
        return dependency_dirs

    def __save_install_log(self):
        tmpdir = self.tmp_install_dir.name
        tmp_install_log = os.path.join(tmpdir, "install.log")
        with open(tmp_install_log, "w") as f:
            f.write(self.install_log)

    def __save_index(self):
        index_location = self.__path_mappings["index"]
        index_dir = os.path.dirname(os.path.abspath(index_location))
        if not os.path.exists(index_dir):
            os.makedirs(index_dir)
        with open(index_location, "w") as f:
            json.dump(self.index, f, indent=2)

    def move_src(self, src, dst):
        if src == "" or not os.path.exists(src) or not os.path.isdir(src):
            raise ValueError("src {} is not directory".format(src))
        if dst == "" or ".." in dst:
            raise ValueError("dst {} is invalid".format(dst))
        # we use cp command here because shutil module is slow,
        # but the behavior of cp command is slightly different between Mac and Linux
        # we use a command like `cp -r <src>/* <dst>/` so the behavior will be the same
        os.system("cp -r {}/* {}/".format(src, dst))
        return

    def setup_tmp_dir(self):
        if self.tmp_install_dir is None or not os.path.exists(self.tmp_install_dir.name):
            self.tmp_install_dir = tempfile.TemporaryDirectory()

    def clean_tmp_dir(self):
        if self.tmp_install_dir is not None and os.path.exists(self.tmp_install_dir.name):
            self.tmp_install_dir.cleanup()
            self.tmp_install_dir = None

    def export_data(self, data, dir, filename):
        if not os.path.exists(dir):
            os.makedirs(dir)
        file = os.path.join(dir, filename)
        logger.debug("export data {} to {}".format(data, file))
        with open(file, "w") as f:
            json.dump(data, f)
        return file

    def get_metafile_in_target(self, type, filepath):
        metafile_path = ""
        files_path = ""
        if type == LoadType.COLLECTION:
            # get manifest.json
            with tarfile.open(name=filepath, mode="r") as tar:
                for info in tar.getmembers():
                    if info.name.endswith(collection_manifest_json):
                        f = tar.extractfile(info)
                        metafile_path = filepath.replace(".tar.gz", "-{}".format(collection_manifest_json))
                        with open(metafile_path, "wb") as c:
                            c.write(f.read())
                    if info.name.endswith(collection_files_json):
                        f = tar.extractfile(info)
                        files_path = filepath.replace(".tar.gz", "-{}".format(collection_files_json))
                        with open(files_path, "wb") as c:
                            c.write(f.read())
        elif type == LoadType.ROLE:
            # get meta/main.yml path
            role_meta_files = safe_glob(
                [
                    os.path.join(filepath, "**", role_meta_main_yml),
                    os.path.join(filepath, "**", role_meta_main_yaml),
                ],
                recursive=True,
            )
            if len(role_meta_files) != 0:
                metafile_path = role_meta_files[0]
        return metafile_path, files_path

    def update_metadata(self, type, metadata_file, target, key, value):
        with open(metadata_file, "r") as f:
            metadata = json.load(f)
        if type == LoadType.COLLECTION:
            metadata_list = metadata.get("collections", [])
        elif type == LoadType.ROLE:
            metadata_list = metadata.get("roles", [])
        else:
            logger.warning("metadata not found: unsupported type {}".format(target))
            return None
        for i, data in enumerate(metadata_list):
            dm = DownloadMetadata(**data)
            if dm.name == target:
                if hasattr(dm, key):
                    setattr(dm, key, value)
                metadata_list[i] = asdict(dm)
                logger.debug("update {} in metadata: {}".format(key, dm))
                if type == LoadType.COLLECTION:
                    metadata["collections"] = metadata_list
                elif type == LoadType.ROLE:
                    metadata["roles"] = metadata_list
                with open(metadata_file, "w") as f:
                    json.dump(metadata, f)
        return

    def update_role_download_src(self, metadata_file, dst_src_dir):
        with open(metadata_file, "r") as f:
            metadata = json.load(f)
        metadata_list = metadata.get("roles", [])
        for i, data in enumerate(metadata_list):
            dm = DownloadMetadata(**data)
            value = "{}/{}".format(dst_src_dir, dm.name)
            key = "download_src_path"
            if hasattr(dm, key):
                setattr(dm, key, value)
            dm.metafile_path, _ = self.get_metafile_in_target(LoadType.ROLE, value)
            dm.author = self.get_author(LoadType.ROLE, dm.metafile_path)
            metadata_list[i] = asdict(dm)
            logger.debug("update {} in metadata: {}".format(key, dm))
        metadata["roles"] = metadata_list
        with open(metadata_file, "w") as f:
            json.dump(metadata, f)
        return

    def get_author(self, type, metafile_path):
        if not os.path.exists(metafile_path):
            logger.warning("invalid file path: {}".format(metafile_path))
            return ""
        if type == LoadType.COLLECTION:
            with open(metafile_path, "r") as f:
                metadata = json.load(f)
            authors = metadata.get("collection_info", {}).get("authors", [])
            return ",".join(authors)
        elif type == LoadType.ROLE:
            with open(metafile_path, "r") as f:
                metadata = yaml.safe_load(f)
            author = metadata.get("galaxy_info", {}).get("author", "")
            return author


def find_ext_dependencies(path):
    collection_meta_files = safe_glob(os.path.join(path, "**", collection_manifest_json), recursive=True)
    if len(collection_meta_files) > 0:
        collection_path_list = [trim_suffix(f, ["/" + collection_manifest_json]) for f in collection_meta_files]
        collection_path_list = remove_subdirectories(collection_path_list)
        return LoadType.COLLECTION, collection_path_list
    role_meta_files = safe_glob(
        [
            os.path.join(path, "**", role_meta_main_yml),
            os.path.join(path, "**", role_meta_main_yaml),
        ],
        recursive=True,
    )
    if len(role_meta_files) > 0:
        role_path_list = [trim_suffix(f, ["/" + role_meta_main_yml, "/" + role_meta_main_yaml]) for f in role_meta_files]
        role_path_list = remove_subdirectories(role_path_list)
        return LoadType.ROLE, role_path_list
    return LoadType.UNKNOWN, []



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from ruamel.yaml import YAML


_yaml = YAML(typ="rt", pure=True)
_yaml.default_flow_style = False
_yaml.preserve_quotes = True
_yaml.allow_duplicate_keys = True


def load(stream: any):
    return _yaml.load(stream)


def dump(data: any):
    output = io.StringIO()
    _yaml.dump(data, output)
    return output.getvalue()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

key_delimiter = ":"
object_delimiter = "#"


class Key:
    def __init__(self, str):
        self.key = str

    def detect_type(self):
        return self.key.split(" ")[0]

    # convert key to tree root name
    def to_name(self):
        _type = self.detect_type()
        if _type == "playbook":
            return os.path.basename(self.key.split(key_delimiter)[-1])
        elif _type == "role":
            return self.key.split(key_delimiter)[-1]


def make_global_key_prefix(collection, role):
    key_prefix = ""
    if collection != "":
        key_prefix = "collection{}{}{}".format(key_delimiter, collection, object_delimiter)
    elif role != "":
        key_prefix = "role{}{}{}".format(key_delimiter, role, object_delimiter)
    return key_prefix


def detect_type(key=""):
    return key.split(" ")[0]


def set_play_key(obj, parent_key="", parent_local_key=""):
    type_str = obj.type
    index_info = "[{}]".format(obj.index)
    _parent_key = parent_key.split(" ")[-1]
    _parent_local_key = parent_local_key.split(" ")[-1]
    global_key = "{} {}{}{}{}{}".format(
        type_str,
        _parent_key,
        object_delimiter,
        type_str,
        key_delimiter,
        index_info,
    )
    local_key = "{} {}{}{}{}{}".format(
        type_str,
        _parent_local_key,
        object_delimiter,
        type_str,
        key_delimiter,
        index_info,
    )
    obj.key = global_key
    obj.local_key = local_key


def set_role_key(obj):
    global_key_prefix = make_global_key_prefix(obj.collection, "")
    global_key = "{} {}{}{}{}".format(obj.type, global_key_prefix, obj.type, key_delimiter, obj.fqcn)
    local_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.defined_in)
    obj.key = global_key
    obj.local_key = local_key


def set_module_key(obj):
    global_key_prefix = make_global_key_prefix(obj.collection, obj.role)
    global_key = "{} {}{}{}{}".format(obj.type, global_key_prefix, obj.type, key_delimiter, obj.fqcn)
    local_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.defined_in)
    obj.key = global_key
    obj.local_key = local_key


def set_collection_key(obj):
    global_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.name)
    local_key = global_key
    obj.key = global_key
    obj.local_key = local_key


def get_obj_type(key):
    idx0 = key.find(" ")
    obj_type = key[:idx0]
    if obj_type in [
        "module",
        "play",
        "playbook",
        "role",
        "collection",
        "task",
        "taskfile",
        "repository",
    ]:
        return obj_type
    else:
        return None


def get_obj_info_by_key(key):
    info = {}
    info["key"] = key
    skip = False

    idx0 = key.find(" ")
    obj_type = key[:idx0]
    info["type"] = obj_type
    skip = idx0 < 0
    if not skip:
        s1 = key[idx0 + 1 :]
        if obj_type == "task" or obj_type == "play":
            idx1 = s1.find(object_delimiter)
            parent_key = s1[:idx1]
            info["parent_key"] = parent_key
            sidx1 = parent_key.find(":")
            parent_type = parent_key[:sidx1]
            info["parent_type"] = parent_type
            parent_name = parent_key[sidx1 + 1 :]
            info["parent_name"] = parent_name
            skip = skip or idx1 < 0
            if not skip:
                s2 = s1[idx1 + 1 :]
                idx2 = s2.find(key_delimiter)
                obj_type = s2[:idx2]
                info["obj_type"] = obj_type
                skip = skip or idx2 < 0
            if not skip:
                obj_key = s2[idx2 + 1 :]
                info["obj_key"] = obj_key
        elif obj_type == "taskfile" or obj_type == "playbook":
            idx1 = s1.find(key_delimiter)
            skip = skip or idx1 < 0
            if not skip:
                parent_type = s1[:idx1]
                info["parent_type"] = parent_type
                s2 = s1[idx1 + 1 :]
                idx2 = s2.find(object_delimiter)
                skip = skip or idx2 < 0
            if not skip:
                parent_name = s2[:idx2]
                info["parent_name"] = parent_name
                s3 = s2[idx2 + 1 :]
                idx3 = s3.find(key_delimiter)
                skip = skip or idx3 < 0
            if not skip:
                defined_in = s3[idx3 + 1 :]
                info["defined_in"] = defined_in
        elif obj_type == "role" or obj_type == "module":
            idx1 = s1.find(key_delimiter)
            parent_type = s1[:idx1]
            info["parent_type"] = parent_type
            skip = skip or idx1 < 0
            if not skip:
                s2 = s1[idx1 + 1 :]
                idx2 = s2.find(object_delimiter)
                parent_name = s2[:idx2]
                info["parent_name"] = parent_name
                skip = skip or idx2 < 0
            if not skip:
                s3 = s2[idx2 + 1 :]
                idx3 = s3.find(key_delimiter)
                skip = skip or idx3 < 0
            if not skip:
                fqcn = s3[idx3 + 1 :]
                info["fqcn"] = fqcn
        elif obj_type == "collection" or obj_type == "repository":
            idx1 = s1.find(key_delimiter)
            skip = skip or idx1 < 0
            if not skip:
                name = s1[idx1 + 1 :]
                info["name"] = name
        else:
            pass

    return info


def set_task_key(obj, parent_key="", parent_local_key=""):
    index_info = "[{}]".format(obj.index)
    _parent_key = parent_key.split(" ")[-1]
    _parent_local_key = parent_local_key.split(" ")[-1]
    global_key = "{} {}{}{}{}{}".format(
        obj.type,
        _parent_key,
        object_delimiter,
        obj.type,
        key_delimiter,
        index_info,
    )
    local_key = "{} {}{}{}{}{}".format(
        obj.type,
        _parent_local_key,
        object_delimiter,
        obj.type,
        key_delimiter,
        index_info,
    )
    obj.key = global_key
    obj.local_key = local_key


def set_taskfile_key(obj):
    global_key_prefix = make_global_key_prefix(obj.collection, obj.role)
    global_key = "{} {}{}{}{}".format(
        obj.type,
        global_key_prefix,
        obj.type,
        key_delimiter,
        obj.defined_in,
    )
    local_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.defined_in)
    obj.key = global_key
    obj.local_key = local_key


def set_playbook_key(obj):
    global_key_prefix = make_global_key_prefix(obj.collection, obj.role)
    global_key = "{} {}{}{}{}".format(
        obj.type,
        global_key_prefix,
        obj.type,
        key_delimiter,
        obj.defined_in,
    )
    local_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.defined_in)
    obj.key = global_key
    obj.local_key = local_key


def set_repository_key(obj):
    global_key = "{} {}{}{}".format(obj.type, obj.type, key_delimiter, obj.name)
    local_key = global_key
    obj.key = global_key
    obj.local_key = local_key


def set_call_object_key(cls_name: str, spec_key: str, caller_key: str):
    parts = spec_key.split(" ", 1)
    caller_only = caller_key.split(" FROM ")[0]
    return "{} {} FROM {}".format(cls_name, parts[1], caller_only)
